# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

"""
Functionality related to downloading objects from S3.
"""
import csv
import datetime
import logging
import os
import threading
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, fields
from itertools import compress
from pathlib import Path
from typing import (
    Any,
    Callable,
    Collection,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Tuple,
    Union,
)

import boto3.session
import botocore.exceptions
from botocore.client import Config
from ego4d.cli import manifest
from ego4d.cli.config import DATASET_FILE_EXTENSIONS, DATASETS_VIDEO, ValidatedConfig
from ego4d.cli.manifest import VideoMetadata
from tqdm import tqdm


__VERSION_ENTRY_FILENAME = "manifest.ver"
SAVE_INTERVAL_S = 15


@dataclass
class VersionEntry:
    uid: str
    version: str
    filename: str


@dataclass
class FileToDownload:
    """Data object that tracks a video to be downloaded to a specific path"""

    filename: str
    download_folder: Path
    uid: str
    s3_bucket: str = None
    s3_object_key: str = None
    s3_object: Any = None
    s3_exists: Optional[bool] = None
    # Size of the file on S3. This should match the file size on disk.
    s3_content_size_bytes: int = 0
    s3_version: str = None
    # Location of the local file downloaded.
    file_path: Optional[Path] = None

    def exists(self) -> bool:
        if not self.s3_object:
            raise RuntimeError(f"No s3_object for exists call: {self.uid}")
        try:
            self.s3_object.load()
            return True
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "404":
                return False
            if e.response["Error"]["Code"] == "403":
                print(f"Boto 403 Exception For exists: {self.uid} | {self.filename}")
                return False
            print(f"Boto Unexpected Exception For exists: {self.uid} | {self.filename}")
            raise

    def file_version_base(self) -> str:
        if not self.filename:
            raise RuntimeError("Invalid filename for file_version")
        base, ext = os.path.splitext(self.filename)
        if ext not in DATASET_FILE_EXTENSIONS:
            logging.warning(
                f"Unexpected file_version extension: {ext} filename: {self.filename}"
            )
        assert "/" not in base
        return base

    def file_version_name(self, version: str) -> str:
        """Creates a file_version_name for a video that is being downloaded"""
        base = self.file_version_base()
        return f"{base}.{version}"

    def file_version_pattern(self) -> str:
        base = self.file_version_base()
        return f"{base}.*"

    def to_version_entry(self) -> VersionEntry:
        return VersionEntry(
            uid=self.uid, version=self.s3_version, filename=self.filename
        )

    @staticmethod
    def create(video: VideoMetadata, download_folder: Path) -> "FileToDownload":
        assert video.filename_base, f"VideoMetadata missing filename_base: {video.uid}"

        x = FileToDownload(
            filename=video.filename_base,
            download_folder=download_folder,
            uid=video.uid,
        )
        if video.s3_bucket and video.s3_object_key:
            x.s3_bucket = video.s3_bucket
            x.s3_object_key = video.s3_object_key
        return x


@dataclass
class VideoOnDisk:
    """
    Data object representing a video that has been downloaded from S3 to the local hard
    disk.
    """

    # Path to the file on disk
    file_path: Path

    # Size of the file on S3. This should match the file size on disk.
    s3_content_size_bytes: int


def create_output_directory(validated_cfg: ValidatedConfig) -> Path:
    """
    Creates a top level download directory if it does not already exist, and returns
    the Path to the download directory.
    """
    # only use the major for the output dir, i.e. v2_1 => v2
    download_path = validated_cfg.output_directory / f"{validated_cfg.out_version_dir}"
    download_path.mkdir(parents=True, exist_ok=True)
    return download_path


def create_download_directory(validated_cfg: ValidatedConfig, dataset: str) -> Path:
    """
    Creates a download directory for a dataset if it does not already exist, and returns
    the Path to the download directory.
    """
    download_path = (
        validated_cfg.output_directory / f"{validated_cfg.out_version_dir}/{dataset}"
    )
    download_path.mkdir(parents=True, exist_ok=True)
    return download_path


# def list_s3_objects(downloads: Union[Collection[FileToDownload], tqdm], s3):
#     """
#     Takes a list of videos to be downloaded (optionally wrapped by tqdm to show a
#     progress bar) and returns a corresponding list of S3.Objects for each video that is
#     to be downloaded.

#     Args:
#         downloads:
#         s3 (S3.ServiceResource):

#     Returns:
#     List[S3.Object]
#     """
#     for d in downloads:
#         if not d.s3_object_key:
#             continue

#     return [s3.Object(d.s3_bucket, d.s3_object_key) for d in downloads if d.s3_object_key]


def info(msg):
    logging.info(msg)
    print(msg)


def filter_already_downloaded(
    downloads: Iterable[FileToDownload],
    version_entries: List[VersionEntry],
    bypass_version_check: bool = False,
    skip_s3_checks: bool = False,
) -> List[FileToDownload]:
    """
    Takes a collection of files that are to be downloaded and a list of the S3.Objects
    corresponding to the files to download and removes any that have already been
    downloaded.
    """

    def already_downloaded(download: FileToDownload) -> bool:
        assert download.filename
        # file_version_name = download.file_version_name(s3_object.version_id)
        # assert file_version_name

        download.s3_exists = (
            skip_s3_checks or download.exists()
        )  # shortcircuits for a faster initial download
        if not download.s3_exists:
            info(
                f"Missing s3 object (ignored for download): {download.uid} | {download.filename}"
            )
            return False

        file_location = download.download_folder / download.filename
        # file_version_location = download.download_folder / file_version_name
        if not file_location.exists():
            logging.info(f"already_downloaded: missing file: {file_location}")
            return False

        if not download.s3_object:
            logging.error(
                f"filter_already_downloaded: invalid s3 object: {download.uid}"
            )
            return False

        if bypass_version_check:
            return True

        version_entry = next(
            (x for x in version_entries if x.uid == download.uid), None
        )
        if not version_entry:
            info(
                f"already_downloaded: no version entry for existing file: {file_location}"
            )
            return False

        s3_version = download.s3_object.version_id
        if version_entry.version != s3_version:
            info(
                f"filter_already_downloaded: mismatched s3 object version: {download.uid} "
                f" {version_entry.version} v {s3_version} "
            )
            return False

        s3_size = download.s3_object.content_length
        # Return true if the size on S3 matches the file size on disk
        if file_location.stat().st_size != s3_size:
            logging.warning(f"already_downloaded=False for file size: {file_location}")
            return False

        return True

    with ThreadPoolExecutor(max_workers=15) as pool:
        to_download = list(
            tqdm(
                pool.map(
                    lambda x: x.s3_object and not already_downloaded(x) and x.s3_exists,
                    downloads,
                ),
                total=len(downloads),
                unit="file",
            )
        )

    n_filtered = len(downloads) - sum(to_download)
    if n_filtered > 0:
        info(f"Filtered {n_filtered}/{len(downloads)} existing videos for download.")
    else:
        info("No existing videos to filter.")

    return list(compress(downloads, to_download))


def download_all(
    downloads: Collection[FileToDownload],
    entries: List[VersionEntry],
    aws_profile_name: str,
    callback: Callable[[int], None] = None,
    save_callback: Callable[[], None] = None,
) -> List[FileToDownload]:
    thread_data = threading.local()
    lock_update = threading.Lock()
    last_save = datetime.datetime.utcnow()

    def initializer():
        config = Config(
            connect_timeout=120,
            retries={"mode": "standard", "max_attempts": 5},
        )
        thread_data.s3 = boto3.session.Session(profile_name=aws_profile_name).resource(
            "s3",
            config=config,
        )

    def download_video(download: FileToDownload):
        nonlocal last_save

        assert download.filename
        assert download.s3_bucket
        assert download.s3_object_key

        obj = thread_data.s3.Object(download.s3_bucket, download.s3_object_key)
        download.s3_version = obj.version_id

        # file_version_name = download.file_version_name(obj.version_id)
        file_path = download.download_folder / download.filename

        # TODO: Can remove for ship
        # Remove any existing version files
        old_version_files = list(
            download.download_folder.glob(download.file_version_pattern())
        )
        for file in old_version_files:
            # file.unlink()
            os.unlink(str(file))

        # Remove the old video file (if it exists)
        if os.path.exists(file_path):
            # file_path.unlink(missing_ok=True)
            os.unlink(str(file_path))

        # Create an empty file with the version name so that it can be tracked later
        # (download.download_folder / file_version_name).touch()

        try:
            obj.download_file(str(file_path), Callback=callback)

            download.file_path = file_path
            download.s3_content_size_bytes = obj.content_length

            with lock_update:
                now = datetime.datetime.utcnow()
                upsert_version(download, entries)
                if save_callback:
                    delta_s = (now - last_save).total_seconds()
                    if delta_s > SAVE_INTERVAL_S:
                        logging.debug("Incremental save..")
                        save_callback()
                        last_save = now

        except Exception as ex:
            logging.exception(f"S3 Download Exception: {ex}")
            download.file_path = None
            download.s3_content_size_bytes = 0

        return download

    # Note: The download operation will be I/O bound by network bandwidth, not compute
    # bound, so a thread pool executor can be used instead of a process pool executor.
    with ThreadPoolExecutor(max_workers=8, initializer=initializer) as pool:
        results = list(pool.map(download_video, downloads))

    return results


def list_videos_for_download(
    cfg: ValidatedConfig, dataset: str, manifest_path: Path
) -> List[VideoMetadata]:
    """
    Takes a user-supplied configuration and a video manifest from S3 then returns a
    list of all videos from the manifest that the user has requested.

    Args:
        cfg:
        manifest_path:
    """
    generator = manifest.list_videos_in_manifest(
        manifest_path, cfg.benchmarks, cfg.universities
    )

    videos = list(generator)

    if cfg.video_uids:
        if any(x != x.lower() for x in cfg.video_uids):
            raise RuntimeError(
                "ERROR: Upper case uids invalid - please sanity check your inputs: {cfg.video_uids}"
            )
        matches = [x for x in videos if x.uid in cfg.video_uids]
        if dataset in DATASETS_VIDEO or dataset.lower() in DATASETS_VIDEO:
            missing = cfg.video_uids - {v.uid for v in videos}
            if missing:
                raise RuntimeError(
                    "The following requested video UIDs could not be found "
                    f"in the manifest for version: '{cfg.version}' and "
                    f"dataset(s): '{cfg.datasets}':\n"
                    f"{missing}"
                )

            videos = matches
        else:
            if matches:
                print(
                    "ERROR: video_uids not supported for non-video datasets: {[x.uid for x in matches]}"
                )

    return videos


def list_corrupt_files(downloads: Collection[FileToDownload]) -> List[FileToDownload]:
    """
    Returns a list of any downloaded files that appear corrupted.
    """
    return [d for d in downloads if _file_is_corrupt(d)]


def _file_is_corrupt(download: FileToDownload):
    if download.file_path and download.file_path.exists():
        return download.file_path.stat().st_size != download.s3_content_size_bytes

    return True


def load_version_file(download_path: Path) -> List[VersionEntry]:
    file_path = download_path / __VERSION_ENTRY_FILENAME
    if not os.path.exists(file_path):
        return []
    with open(file_path) as f:
        reader = csv.DictReader(f)
        data = list(reader)
        entries = [
            VersionEntry(row["uid"], row["version"], row["filename"]) for row in data
        ]
    return entries


def save_version_file(entries: List[VersionEntry], download_path: Path):
    file_path = download_path / __VERSION_ENTRY_FILENAME
    with open(file_path, "w", newline="", encoding="utf-8") as f:
        fieldnames = [field.name for field in fields(VersionEntry)]
        writer = csv.DictWriter(f, fieldnames)
        writer.writeheader()
        writer.writerows(
            {"uid": x.uid, "version": x.version, "filename": x.filename}
            for x in entries
        )


def upsert_version(download: FileToDownload, entries: List[VersionEntry]):
    assert download and download.uid

    if download.s3_content_size_bytes == 0:
        # download failed, just ignore?
        logging.warning("upsert_version: ignoring 0 byte download: " + download.uid)
        return

    matches = [x for x in entries if x.uid == download.uid]
    if len(matches) == 0:
        entries.append(download.to_version_entry())
    else:
        assert (
            len(matches) == 1
        ), f"Multiple version entries for uid invalid: {download.uid}"
        entry = matches[0]
        if download.s3_version:
            entry.version = download.s3_version
        if download.filename != entry.filename:
            logging.error(
                f"Suspect version info: Filename ({download.filename}) changed for existing entry: {entry.filename}"
            )
            entry.filename = download.filename



import functools
import json
import os
from typing import Any, Dict, List, Tuple

import submitit.helpers as sh

import torch
from ego4d.research.clep.config import (
    EgoPreprocessFeatureConfig,
    EgoPreprocessNarrConfig,
    InputConfig,
    TrainConfig,
)
from ego4d.research.clep.preprocess.common import get_language_model
from ego4d.research.common import batch_it, create_executor
from ego4d.research.dataset import save_features_to_hdf5

from tqdm.auto import tqdm


def preprocess_ego_narrations(
    config: TrainConfig, narr_config: EgoPreprocessNarrConfig
):
    out_dir = config.pre_config.root_dir
    os.makedirs(out_dir, exist_ok=True)

    narr_od = os.path.join(out_dir, narr_config.narration_out_dir)
    os.makedirs(narr_od, exist_ok=True)

    narrs = get_narrations(narr_config)
    narrs = filter_narrations(narrs, config.input_config, narr_config)

    print("Transforming text...")
    narrs_with_idx = list(
        enumerate(
            [
                (remove_tags(txt), sub_tagged_tokens(txt), uid, txt, ts)
                for uid, txt, ts in narrs
            ]
        )
    )
    narrs_with_idx = narrs_with_idx[0 : narr_config.limit]

    batches = batch_it(narrs_with_idx, narr_config.num_narrs_per_machine)
    print(f"Running txt through transformer with {len(batches)} machines")
    print(f"Num narrs = {len(narrs_with_idx)}", flush=True)

    metas = []
    executor = create_executor(config.pre_config.slurm_config, len(batches))
    jobs = executor.map_array(
        functools.partial(
            _map_narrs_on_machine,
            config=config,
            narr_od=narr_od,
        ),
        batches,
    )
    print("Jobs", jobs, flush=True)

    metas = []
    for j in tqdm(sh.as_completed(jobs), total=len(jobs)):
        res = j.result()
        metas.extend(res)

    print("Saving metadata")
    m_op = os.path.join(out_dir, narr_config.metadata_out_path)
    torch.save(metas, m_op)


def preprocess_ego_features(
    feature_path: str,
    config: TrainConfig,
    pre_feature: EgoPreprocessFeatureConfig,
):
    out_dir = config.pre_config.root_dir
    os.makedirs(out_dir, exist_ok=True)

    meta = json.load(open(config.input_config.metadata_path))
    video_uids = [x["video_uid"] for x in meta["videos"]]

    out_path = os.path.join(config.pre_config.root_dir, pre_feature.hdf5_path)
    print("=>", out_path, flush=True)
    save_features_to_hdf5(
        video_uids=video_uids,
        feature_dir=feature_path,
        out_path=out_path,
    )


def _map_narrs_on_machine(
    narrs: List[Tuple[int, Tuple[str, str, str, str, float]]],
    config: TrainConfig,
    narr_od: str,
) -> List[Dict[str, Any]]:
    model = get_language_model(config)

    batches = batch_it(narrs, config.batch_size)

    metas = []
    for batch in tqdm(batches):
        fvs = model.encode(
            [x for _, (_, x, _, _, _) in batch],
            device="cuda" if config.accelerator == "gpu" else config.accelerator,
            show_progress_bar=False,
        )
        fvs_without_tags = model.encode(
            [x for _, (x, _, _, _, _) in batch],
            device="cuda" if config.accelerator == "gpu" else config.accelerator,
            show_progress_bar=False,
        )

        for fv_no_tag, fv, (idx, (no_tag_txt, post_txt, uid, txt, ts)) in zip(
            fvs_without_tags, fvs, batch
        ):
            od = os.path.join(narr_od, uid)
            os.makedirs(od, exist_ok=True)
            path_to_encode = os.path.join(od, f"{idx}.pt")
            torch.save(
                {
                    "fv": fv,
                    "fv_no_tag": fv_no_tag,
                },
                path_to_encode,
            )
            metas.append(
                {
                    "uid": uid,
                    "txt": txt,
                    "ts": ts,
                    "idx": idx,
                    "post_txt": post_txt,
                    "no_tag_txt": no_tag_txt,
                }
            )
    return metas


def remove_tags(text: str) -> str:
    text = text.replace("# C", "")
    text = text.replace("#C", "")
    text = text.replace("C", "")
    text = text.replace("#O", "")
    text = text.replace("O", "")
    text = text.replace("#unsure", "")
    return text


def sub_tagged_tokens(text: str) -> str:
    text = text.replace("#C", "Camera wearer")
    text = text.replace("#O", "Other person")
    text = text.replace("#unsure", "something")
    return text


def filter_narrations(
    narrations, config: InputConfig, pre_config: EgoPreprocessNarrConfig
):
    meta = json.load(open(config.metadata_path))
    val_set_uids = [
        vid["video_uid"]
        for vid in meta["videos"]
        if len({vid["split_em"], vid["split_fho"], vid["split_av"]} & {"val", "test"})
        > 0
    ]

    num_val_filtered = 0
    num_txt_filtered = 0
    ret = []
    for uid, txt, ts in tqdm(narrations):
        if uid in val_set_uids:
            num_val_filtered += 1
            continue
        if len(txt.split(" ")) < pre_config.min_words:
            num_txt_filtered += 1
            continue
        ret.append((uid, txt, ts))
    print(f"Narrations filtered from {len(narrations)} -> {len(ret)}")
    print(
        f"""
    Val Filtered = {num_val_filtered} = {num_val_filtered/len(narrations):.2%}
    Txt Filtered = {num_txt_filtered} = {num_txt_filtered/len(narrations):.2%}

    """
    )
    return ret


def get_narrations(
    config: EgoPreprocessNarrConfig,
) -> List[Tuple[str, str, float, str]]:
    narration_json = json.load(open(config.narration_json_path))
    uid_subset = set(narration_json.keys())
    narrations = [
        (uid, data["narration_text"], data["timestamp_sec"])
        for uid in uid_subset
        for data in narration_json[uid].get("narration_pass_1", {"narrations": []})[
            "narrations"
        ]
    ]
    narrations += [
        (uid, data["narration_text"], data["timestamp_sec"])
        for uid in uid_subset
        for data in narration_json[uid].get("narration_pass_2", {"narrations": []})[
            "narrations"
        ]
    ]
    narrations.sort(key=lambda x: (x[0], x[-1]))
    return narrations



# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.

from dataclasses import dataclass
from typing import Tuple

import torch
from ego4d.features.config import BaseModelConfig, InferenceConfig
from pytorchvideo.transforms import ApplyTransformToKey, ShortSideScale
from torch.nn import Identity, Module
from torchvision.transforms import Compose, Lambda
from torchvision.transforms._transforms_video import CenterCropVideo, NormalizeVideo


@dataclass
class ModelConfig(BaseModelConfig):
    model_name: str = "omnivore_swinB"
    input_type: str = "video"
    side_size: int = 256
    crop_size: int = 224
    mean: Tuple[float] = (0.485, 0.456, 0.406)
    std: Tuple[float] = (0.229, 0.224, 0.225)


class WrapModel(Module):
    def __init__(self, model: Module, input_type: str):
        super().__init__()
        self.model = model
        self.input_type = input_type

    def forward(self, x) -> torch.Tensor:
        return self.model(x["video"], input_type=self.input_type)


def load_model(
    inference_config: InferenceConfig,
    config: ModelConfig,
    patch_final_layer: bool = True,
) -> Module:
    model = torch.hub.load("facebookresearch/omnivore", model=config.model_name)

    if patch_final_layer:
        model.heads.image = Identity()
        model.heads.video = Identity()
        model.heads.rgbd = Identity()

    # Set to GPU or CPU
    model = WrapModel(model, config.input_type)
    model = model.eval()
    model = model.to(inference_config.device)
    return model


def norm_pixels(x):
    return x / 255.0


def get_transform(inference_config: InferenceConfig, config: ModelConfig):
    if config.input_type == "video":
        transforms = [
            Lambda(norm_pixels),
            NormalizeVideo(config.mean, config.std),
            ShortSideScale(size=config.side_size),
            CenterCropVideo(config.crop_size),
        ]
    else:
        assert inference_config.frame_window == 1
        transforms = [
            Lambda(norm_pixels),
            NormalizeVideo(config.mean, config.std),
            ShortSideScale(size=config.side_size),
            CenterCropVideo(config.crop_size),
        ]

    return ApplyTransformToKey(
        key="video",
        transform=Compose(transforms),
    )



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

"""
Functionality for parsing AWS S3 paths.

References:
    https://docs.aws.amazon.com/AmazonS3/latest/userguide/access-bucket-intro.html#accessing-a-bucket-using-S3-format
    https://docs.aws.amazon.com/AmazonS3/latest/userguide/bucketnamingrules.html
    https://docs.aws.amazon.com/AmazonS3/latest/userguide/object-keys.html

"""
import re
from typing import Tuple


__S3_PATH_REGEX = re.compile(r"^s3://(?P<bucket>[^/]*)/(?P<key>.*)$")


def bucket_and_key_from_path(path: str) -> Tuple[str, str]:
    """
    Takes an S3 path (i.e. s3://<bucket-name>/<object-key>) and returns the bucket and
    object key as a string tuple.

    Notes:
        This does not validate whether or not the bucket and key satisfy the naming and
        character requirements imposed by AWS (e.g character limits on bucket names).
    """
    match = __S3_PATH_REGEX.match(path)
    return match.group("bucket"), match.group("key")



import copy
import gc
import math
import os
import time

import hydra
import submitit
import torch

import torch.nn.functional as F
from ego4d.research.clep.config import TrainConfig
from ego4d.research.clep.dataset import (
    CCDset,
    create_data_loader,
    create_ego_charades_dset,
    create_kinetics_dset,
    Ego4DCLEP,
)
from ego4d.research.clep.model import EgoLangaugeAssociation

from ego4d.research.clep.val import eval_classification, eval_multi_class_classification

from pytorch_lightning.lite import LightningLite

from torch.utils.tensorboard import SummaryWriter

from tqdm.auto import tqdm


class Lite(LightningLite):
    def my_setup(self, config: TrainConfig):
        self.model = EgoLangaugeAssociation(config.model_config)
        self.config = config
        self.val_config = copy.deepcopy(config)
        self.val_config.batch_size = 1

        self.val_bs_config = copy.deepcopy(config)
        self.val_bs_config.batch_size = 128

    def run(self):
        named_parameters = list(self.model.named_parameters())

        # https://github.com/mlfoundations/open_clip/blob/main/src/training/main.py#L163
        exclude = (
            lambda n, p: p.ndim < 2
            or "bn" in n
            or "ln" in n
            or "bias" in n
            or "logit_scale" in n
        )
        include = lambda n, p: not exclude(n, p)
        gain_or_bias_params = [
            p for n, p in named_parameters if exclude(n, p) and p.requires_grad
        ]
        rest_params = [
            p for n, p in named_parameters if include(n, p) and p.requires_grad
        ]

        self.optimizer = torch.optim.AdamW(
            [
                {"params": gain_or_bias_params, "weight_decay": 0.0},
                {"params": rest_params, "weight_decay": self.config.wd},
            ],
            lr=self.config.lr,
            betas=(self.config.beta1, self.config.beta2),
            eps=self.config.eps,
        )

        model, optimizer = self.setup(self.model, self.optimizer)

        print("Config=")
        print(self.config, flush=True)

        dsets = []
        for d in self.config.input_config.dsets_to_use:
            if d == "ego4d":
                dsets.append(Ego4DCLEP(config=self.config))
            elif d == "cc":
                dsets.append(CCDset(config=self.config))
            else:
                raise AssertionError(f"invalid dataset {d}")

        dset = torch.utils.data.ConcatDataset(dsets)

        dataloader = create_data_loader(dset, self.config)
        dataloader = self.setup_dataloaders(dataloader)

        log_dir = os.path.join(self.config.tb_log_dir, self.config.tb_log_name)
        os.makedirs(log_dir, exist_ok=True)
        os.makedirs(self.config.checkpoint_dir, exist_ok=True)
        writer = SummaryWriter(log_dir=log_dir)

        model.train()
        step = 0
        num_examples = 0

        max_steps = 2 * len(dataloader)  # TODO configure
        scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
            optimizer, T_max=max_steps, eta_min=0, last_epoch=-1
        )

        bce_loss = torch.nn.BCEWithLogitsLoss()

        best_model = None
        loss = torch.tensor(-1.0)

        for epoch in range(self.config.num_epochs):
            print("Epoch:", epoch)
            for batch in tqdm(dataloader, total=len(dataloader)):
                if step % self.config.eval_per_iter == 0 and (
                    step > 0 or self.config.eval_init
                ):
                    print()
                    if step > 0:
                        print("Loss=", loss.cpu().item(), flush=True)
                    print(f"Eval {step} - {self.config.eval_per_iter}", flush=True)
                    kv_pairs = self.run_eval()
                    for k, v in kv_pairs.items():
                        print(k, v)
                        writer.add_scalar(k, v, num_examples)
                    m_value = kv_pairs[self.config.checkpoint_metric]
                    if best_model is None or best_model[0] < m_value:
                        prev_metric = (
                            f"{best_model[0]:.3f}" if best_model is not None else None
                        )
                        print(f"Saving {prev_metric} -> {m_value:.3f}", flush=True)
                        path = f"{self.config.tb_log_name}_{m_value:.3f}.pt"
                        abs_path = os.path.join(self.config.checkpoint_dir, path)
                        torch.save(model, abs_path)
                        # self.save(model.state_dict(), abs_path)
                        if best_model is not None:
                            os.remove(best_model[1])
                        best_model = (m_value, abs_path)
                    print()

                optimizer.zero_grad()
                v_f, t_f, logit_scale = self.model(batch)

                # NOTE:
                # For ego4d I trim away tags for similarity as these tags can
                # make up a significant percentage of the text and hence be
                # quite dramatic since SentenceTransformers perform a pool over
                # the tokens
                txt = batch["text_no_tag"]

                device = txt.device
                with torch.no_grad():
                    if self.config.use_soft_loss is not None:
                        simm = txt @ txt.t()
                        simm = (simm + 1) / 2
                        if self.config.use_soft_loss:
                            label = simm
                        else:
                            label = torch.ones_like(simm, dtype=torch.float)
                        # TODO: nn.Parameter for the threshold?
                        label[simm < self.config.soft_loss_threshold] = 0.0
                        pos_ex_prop = label.sum() / (label.shape[0] * label.shape[0])
                    else:
                        label = torch.arange(txt.shape[0], device=device)
                        pos_ex_prop = None

                if self.config.norm_logits:
                    v_f = F.normalize(v_f, dim=-1)
                    t_f = F.normalize(t_f, dim=-1)

                if self.config.use_logit_scale:
                    vid2txt = logit_scale * v_f @ t_f.T
                    txt2vid = logit_scale * t_f @ v_f.T
                else:
                    vid2txt = v_f @ t_f.T
                    txt2vid = t_f @ v_f.T

                if self.config.use_bce:
                    assert self.config.use_bce != False  # noqa
                    loss = (bce_loss(vid2txt, label) + bce_loss(txt2vid, label)) / 2.0
                else:
                    loss = (
                        F.cross_entropy(vid2txt, label)
                        + F.cross_entropy(txt2vid, label)
                    ) / 2.0

                self.backward(loss)  # instead of loss.backward()
                optimizer.step()
                scheduler.step()

                # https://github.com/mlfoundations/open_clip/blob/main/src/training/train.py#L100
                with torch.no_grad():
                    model.module.logit_scale.clamp_(0, math.log(100))

                num_examples += batch["video"].shape[0]

                writer.add_scalar("Loss/train", loss.detach().cpu(), num_examples)
                writer.add_scalar(
                    "logit_scale", model.module.logit_scale.detach().cpu(), num_examples
                )
                writer.add_scalar("lr", optimizer.param_groups[0]["lr"], num_examples)
                if pos_ex_prop is not None:
                    writer.add_scalar("pos_ex_prop", pos_ex_prop, num_examples)
                writer.flush()

                if step % 100 == 0:
                    gc.collect()

                step += 1

    def _run_ego_charades(self, ego_only=True, use_ego_sent=True):
        dset, sents = create_ego_charades_dset(
            self.config,
            use_ego_sent=use_ego_sent,
            ego_only=ego_only,
        )
        val_loader = create_data_loader(dset, self.val_config)
        val_loader = self.setup_dataloaders(val_loader)

        classifier = F.normalize(
            self.model.text_proj(sents.to(self.device)).t(),
            dim=-1,
        )
        res = eval_multi_class_classification(
            self.model.visual_proj, classifier, val_loader
        )
        ret = {}
        for k, v in res.items():
            char_name = "Char_"
            if ego_only is not None:
                char_name += f"{ego_only:d}Ego_"
            else:
                char_name += "All_"

            if use_ego_sent is not None:
                char_name += f"{use_ego_sent:d}EgoSent"
            else:
                char_name += "Labels"
            ret[f"Val/{char_name}/{k}"] = v
        return ret

    def _run_eval_kinetics(self):
        dset, sents = create_kinetics_dset(self.config)
        val_loader = create_data_loader(dset, self.val_bs_config, torch.mean)
        val_loader = self.setup_dataloaders(val_loader)

        classifier = F.normalize(
            self.model.text_proj(sents.to(self.device)).t(),
            dim=-1,
        )
        res = eval_classification(
            self.model.visual_proj, classifier, val_loader, avg_logits=False
        )
        return {
            "Val/Kinetics/acc1": res["acc1"],
            "Val/Kinetics/acc5": res["acc5"],
        }

    def run_eval(self):
        self.model.eval()
        result = {}
        result.update(self._run_eval_kinetics())
        result.update(self._run_ego_charades(ego_only=None, use_ego_sent=None))
        result.update(self._run_ego_charades(ego_only=None, use_ego_sent=True))
        result.update(self._run_ego_charades(ego_only=False, use_ego_sent=None))
        result.update(self._run_ego_charades(ego_only=True, use_ego_sent=None))
        result.update(self._run_ego_charades(ego_only=True, use_ego_sent=False))
        result.update(self._run_ego_charades(ego_only=True, use_ego_sent=True))
        self.model.train()
        return result


def run_train(config: TrainConfig):
    lite = Lite(accelerator=config.accelerator, devices=config.devices)
    lite.my_setup(config)
    lite.run()


@hydra.main(config_path="configs", config_name=None)
def train_model(config: TrainConfig):
    print(config)
    if not config.run_locally:
        executor = submitit.AutoExecutor(folder=config.slurm_log_folder)
        executor.update_parameters(
            timeout_min=1200,
            constraint="volta",
            slurm_partition="pixar",
            gpus_per_node=1,
            cpus_per_task=10,
        )
        job_id = executor.submit(run_train, config)
        print(job_id.job_id)
        # _ = job_id.result()
    else:
        run_train(config)


if __name__ == "__main__":
    train_model()  # pyre-ignore



# pyre-strict
import json
import os
from dataclasses import dataclass
from fractions import Fraction
from typing import Iterator, List, Optional

import av
import numpy as np
import torch
from iopath.common.file_io import PathManager


@dataclass
class VideoChunk:
    """
    Normalized from 0-255 (uint8). Shape == (t, c, h, w)
    """

    video_frames: np.array

    """
    Not normalized from 0-1, shape yet to be normalized
    """
    audio_frames: Optional[np.array]

    video_frames_pts: List[int]
    audio_frames_pts: Optional[List[int]]

    video_frames_sec: List[Fraction]
    audio_frames_sec: Optional[List[Fraction]]

    video_timebase: Fraction
    audio_timebase: Fraction
    audio_sample_rate: int
    video_fps: Fraction


def split_av_frames(
    video_path: str,
    window_size_sec: int = 5,
    subsample_n_frames: Optional[int] = None,
    limit: int = -1,
) -> Iterator[VideoChunk]:
    with av.open(video_path) as container:
        has_audio = len(container.streams.audio) > 0
        has_audio = False

        streams_to_decode = {"video": 0}
        video_tb = container.streams.video[0].time_base
        video_fps = container.streams.video[0].framerate
        audio_tb = None
        audio_sample_rate = None
        if has_audio:
            streams_to_decode["audio"] = 0
            audio_tb = container.streams.audio[0].time_base
            audio_sample_rate = container.streams.audio[0].sample_rate

        vf_buffer = {}
        af_buffer = {}
        curr_window_start_sec = 0
        curr_window_end_sec = curr_window_start_sec + window_size_sec
        i = 0
        for frame in container.decode(**streams_to_decode):
            t_pts_sec = frame.pts * frame.time_base
            if t_pts_sec >= curr_window_end_sec:
                i += 1
                curr_window_start_sec = curr_window_start_sec + window_size_sec
                curr_window_end_sec = curr_window_end_sec + window_size_sec
                if len(vf_buffer) > 0 or len(af_buffer) > 0:
                    vf_frames_sorted = sorted(vf_buffer.items(), key=lambda x: x[0])
                    af_frames_sorted = sorted(af_buffer.items(), key=lambda x: x[0])
                    vf_frames_np = np.stack(
                        [x.to_ndarray(format="rgb24") for _, x in vf_frames_sorted]
                    ).transpose(
                        0, 3, 1, 2
                    )  # THWC -> TCHW
                    assert vf_frames_np.dtype == np.uint8
                    indices = None
                    if subsample_n_frames is not None:
                        # ref: https://fburl.com/code/a8viadi2
                        t = vf_frames_np.shape[0]
                        indices = torch.linspace(0, t - 1, subsample_n_frames)
                        indices = torch.clamp(indices, 0, t - 1).long()
                        vf_frames_np = vf_frames_np[indices]
                        indices = set(indices.tolist())

                    # NOTE: do we want to normalize audio from 0-1 as well?
                    af_frames_np = (
                        np.stack([x.to_ndarray() for _, x in af_frames_sorted])
                        if len(af_frames_sorted) > 0
                        else None
                    )
                    vf_frames_pts = [
                        x[0]
                        for i, x in enumerate(vf_frames_sorted)
                        if indices is None or i in indices
                    ]
                    af_frames_pts = (
                        [x[0] for x in vf_frames_sorted]
                        if len(af_frames_sorted) > 0
                        else None
                    )
                    # pyre-ignore
                    yield VideoChunk(
                        video_frames=vf_frames_np,
                        audio_frames=af_frames_np,
                        video_frames_pts=vf_frames_pts,
                        audio_frames_pts=af_frames_pts,
                        video_frames_sec=[x * video_tb for x in vf_frames_pts],
                        audio_frames_sec=(
                            [x * audio_tb for x in af_frames_pts]
                            if af_frames_pts is not None
                            else None
                        ),
                        video_timebase=video_tb,
                        audio_timebase=audio_tb,
                        video_fps=video_fps,
                        audio_sample_rate=audio_sample_rate,
                    )
                vf_buffer = {}
                af_buffer = {}
            else:
                if t_pts_sec < curr_window_start_sec:
                    raise AssertionError("packets from previous window")

            if isinstance(frame, av.AudioFrame):
                af_buffer[frame.pts] = frame
            else:
                assert isinstance(frame, av.VideoFrame)
                vf_buffer[frame.pts] = frame

            if limit > 0 and i >= limit:
                break


def save_chunk(chunk: VideoChunk, out_dir: str, pathmgr: Optional[PathManager]) -> None:
    if pathmgr is None:
        pathmgr = PathManager()

    has_audio = chunk.audio_frames is not None
    is_audio_mono = None
    if has_audio:
        is_audio_mono = (
            len(chunk.audio_frames.shape) == 1 or chunk.audio_frames.shape[1] == 1
        )
        audio_path = os.path.join(out_dir, "audio.npy")

    video_path = os.path.join(out_dir, "video.npy")
    metadata_path = os.path.join(out_dir, "metadata.json")

    metadata = {
        "video_frame_pts": chunk.video_frames_pts,
        "audio_frame_pts": chunk.audio_frames_pts,
        "video_frame_sec": [float(x) for x in chunk.video_frames_sec],
        "audio_frame_sec": (
            [float(x) for x in chunk.audio_frames_sec] if has_audio else None
        ),
        "video_timebase_numerator": chunk.video_timebase.numerator,
        "video_timebase_denom": chunk.video_timebase.denominator,
        "audio_timebase_numerator": (
            chunk.audio_timebase.numerator if chunk.audio_timebase is not None else None
        ),
        "audio_timebase_denom": (
            chunk.audio_timebase.denominator
            if chunk.audio_timebase is not None
            else None
        ),
        "has_audio": has_audio,
        "num_frames": chunk.video_frames.shape[0],
        "num_audio_frames": chunk.audio_frames.shape[0] if has_audio else 0,
        "frame_height": chunk.video_frames.shape[2],
        "frame_width": chunk.video_frames.shape[3],
        "is_audio_mono": is_audio_mono,
        "video_fps_numeratator": chunk.video_fps.numerator,
        "video_fps_denom": chunk.video_fps.denominator,
        "audio_sample_rate": chunk.audio_sample_rate,
    }
    json.dump(metadata, pathmgr.open(metadata_path, "w"), indent=2)

    # TODO: is torch.save better?
    np.save(pathmgr.open(video_path, "wb"), chunk.video_frames)
    if has_audio:
        np.save(pathmgr.open(audio_path, "wb"), chunk.audio_frames)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

"""
A data object for storing user options, along with utilities for parsing input options
from command line flags and a configuration file.
"""
import argparse
import json
from pathlib import Path
from typing import List, NamedTuple, Set, Union

import boto3.session
from botocore.exceptions import ProfileNotFound
from ego4d.cli.universities import UNIV_TO_BUCKET


VERSION_DEFAULT = "v2_1"
DATASET_PRIMARY = "full_scale"
DATASETS_VIDEO = ["full_scale", "clips", "components/videos", "video_540ss"]
DATASETS_FILE = [
    "annotations",
    "viz",
    "imu",
    "gaze",
    "3d",
    "3d_scan_keypoints",
    "3d_scans",
    "av_models",
    "social_test",
    "vq2d_models",
    "vq2d_detections",
    "sta_models",
    "lta_models",
    "slowfast8x8_r101_k400",
    "egotracks",
    "paco_annotations",
    "paco_frames",
]
DATASETS_ALL = DATASETS_VIDEO + DATASETS_FILE
DATASET_FILE_EXTENSIONS = [
    ".mp4",
    ".json",
    ".jsonl",
    ".jpg",
    ".txt",
    ".csv",
    ".pt",
    ".tar",
    ".tgz",
    ".ckpt",
    ".pth",
    ".zip",
]


class ValidatedConfig(NamedTuple):
    """
    Data object that stores validated user-supplied configuration options for a video
    download operation.
    """

    output_directory: Path
    assume_yes: bool
    version: str
    datasets: Set[str]
    benchmarks: Set[str]
    aws_profile_name: str
    metadata: bool
    manifest: bool
    bypass_version_check: bool
    skip_s3_checks: bool
    video_uids: Set[str]
    universities: Set[str]
    annotations: Union[bool, Set[str]]
    list_datasets: bool
    manifest_override_path: str

    @property
    def out_version_dir(self) -> str:
        base_version = self.version.split("_")[0]
        return base_version


class Config(NamedTuple):
    """
    Data object that stores the user-supplied configuration options for a video download
    operation.
    """

    output_directory: str = None
    assume_yes: bool = False
    version: str = VERSION_DEFAULT
    datasets: List[str] = []
    benchmarks: Set[str] = []
    aws_profile_name: str = "default"
    metadata: bool = False
    manifest: bool = False
    bypass_version_check: bool = False
    skip_s3_checks: bool = False
    video_uids: List[str] = []
    universities: List[str] = []
    annotations: Union[bool, List[str]] = True
    list_datasets: bool = False
    manifest_override_path: str = None


def validate_config(cfg: Config) -> ValidatedConfig:
    """

    Args:
        cfg: A user-supplied configuration for the download operation
    Returns:
        A ValidatedConfig if all user-supplied options appear valid
    """

    try:
        boto3.session.Session(profile_name=cfg.aws_profile_name)
    except ProfileNotFound:
        raise RuntimeError(f"Could not find AWS profile '{cfg.aws_profile_name}'.")

    return ValidatedConfig(
        output_directory=(
            Path(cfg.output_directory).expanduser() if cfg.output_directory else None
        ),
        version=cfg.version,
        datasets=set(cfg.datasets),
        benchmarks=set(cfg.benchmarks),
        aws_profile_name=cfg.aws_profile_name,
        manifest=cfg.manifest,
        metadata=cfg.metadata,
        video_uids=set(cfg.video_uids) if cfg.video_uids else {},
        universities=set(cfg.universities) if cfg.universities else {},
        assume_yes=bool(cfg.assume_yes),
        bypass_version_check=cfg.bypass_version_check,
        skip_s3_checks=cfg.skip_s3_checks,
        annotations=cfg.annotations if isinstance(cfg.annotations, bool) else set(),
        list_datasets=cfg.list_datasets,
        manifest_override_path=cfg.manifest_override_path,
    )


def config_from_args(args=None) -> Config:
    """
    Parses command line flags and returns a Config object with corresponding values from
    the flags.
    """
    # Parser for a configuration file
    json_parser = argparse.ArgumentParser(add_help=False)

    json_parser.add_argument(
        "--config_path",
        type=Path,
        help="Local path to a config JSON file. If specified, the flags will be read "
        "from this file instead of the command line.",
    )

    json_args, remaining = json_parser.parse_known_args(args=args)

    # Parser for command line flags other than the configuration file
    flag_parser = argparse.ArgumentParser(
        description="Command line tool to download Ego4D datasets from Amazon S3",
        add_help=True,
    )

    flag_parser.add_argument(
        "-o",
        "--output_directory",
        help="A local path where the downloaded files and metadata will be stored",
    )
    flag_parser.add_argument(
        "--datasets",
        nargs="+",
        help="The datasets to download: 'full_scale', 'annotations', etc."
        "\nThe datasets will be stored in folders in the output directory with the name "
        "of the dataset (e.g. output_dir/full_scale/).",
        default=["annotations"],
    )
    flag_parser.add_argument(
        "--benchmarks",
        nargs="*",
        help="The benchmarks to download dataset subsets of: 'EM', 'FHO', 'AV'",
    )
    flag_parser.add_argument(
        "--annotations",
        nargs="*",
        help="(Currently unsupported) If passed without arguments, downloads all annotations.  "
        "Otherwise, a list of specific annotations to pass, e.g. narration, fho, moments, vq, nlq, av",
    )
    flag_parser.add_argument(
        "--metadata",
        dest="metadata",
        action="store_true",
        help="Download the primary ego4d.json to the folder root.  (Default: True)",
    )
    flag_parser.add_argument(
        "--no-metadata",
        dest="metadata",
        action="store_false",
        help="Bypass downloading the primary ego4d.json.",
    )
    flag_parser.set_defaults(metadata=True)
    flag_parser.add_argument(
        "--manifest",
        const=True,
        action="store_const",
        help="Downloads the video manifest. (True by default, only relevant if you want only the manifest.)",
    )
    # flag_parser.add_argument(
    #     "--viz",
    #     const=True,
    #     action="store_const",
    #     help="Downloads the local visualization dataset. (Convenience option equivalent to including viz in datasets.)",
    # )
    flag_parser.add_argument(
        "--bypass-existing",
        const=True,
        action="store_const",
        dest="bypass_version_check",
        help="Bypass existing files without checking file versions/sizes.",
    )
    flag_parser.add_argument(
        "--skip-s3-checks",
        const=True,
        action="store_const",
        dest="skip_s3_checks",
        help="Skips initial check that files are present on S3 before downloading. Download size estimates are disabled with this flag.",
    )
    flag_parser.add_argument(
        "--version",
        help="A version identifier - i.e. 'v1', 'v2', 'v2_1'",
        default=VERSION_DEFAULT,
    )
    flag_parser.add_argument(
        "--aws_profile_name",
        help="Defaults to 'default'. Specifies the AWS profile name from "
        "~/.aws/credentials to use for the download",
        default="default",
    )
    flag_parser.add_argument(
        "--universities",
        nargs="+",
        choices=UNIV_TO_BUCKET.keys(),
        help="List of university IDs. If specified, only UIDs from the S3 buckets "
        "belonging to the listed universities will be downloaded. A full list of "
        "university IDs can be found in the ego4d/cli/universities.py file.",
    )
    flag_parser.add_argument(
        "-y",
        "--yes",
        default=False,
        const=True,
        action="store_const",
        dest="assume_yes",
        help="If this flag is set, then the CLI will not show a prompt asking the user "
        "to confirm the download. This is so that the tool can be used as part of "
        "shell scripts.",
    )
    video_uid_group = flag_parser.add_mutually_exclusive_group()
    video_uid_group.add_argument(
        "--video_uids",
        nargs="+",
        help="List of video UIDs to be downloaded. If not specified, all relevant UIDs "
        "will be downloaded.",
    )
    video_uid_group.add_argument(
        "--video_uid_file",
        help="Path to a whitespace delimited file that contains a list of UIDs. "
        "Mutually exclusive with the video_uids flag.",
    )
    flag_parser.add_argument(
        "--manifest-override-path",
        help="Override the manifest with the specified path.  Only relevant if exactly "
        "1 dataset is specified",
    )

    flag_parser.add_argument(
        "--list-datasets",
        dest="list_datasets",
        action="store_true",
        help="List the available datasets",
    )

    # Use the values in the config file, but set them as defaults to flag_parser so they
    # can be overridden by command line flags
    if json_args.config_path:
        with open(json_args.config_path.expanduser()) as f:
            config_contents = json.load(f)
            flag_parser.set_defaults(**config_contents)

    parsed_args = flag_parser.parse_args(remaining)
    if parsed_args.datasets is None and parsed_args.annotations is None:
        raise RuntimeError(
            f"Please specify either datasets, annotations or manifest for a minimal download set."
        )

    unknown_datasets = [x for x in parsed_args.datasets if x not in DATASETS_ALL]
    if unknown_datasets:
        print(
            f"Warning: Non-standard Dataset Specfied (Allowed, will attempt download): {unknown_datasets}"
        )

    # if parsed_args.viz:
    #     if parsed_args.datasets:
    #         if "viz" not in parsed_args.datasets:
    #             print("Adding viz to datasets..")
    #             parsed_args.datasets.append("viz")
    #     else:
    #         parsed_args.datasets = ["viz"]
    #     del parsed_args.viz

    help_cmd = False
    if parsed_args.list_datasets:
        help_cmd = True

    flags = {k: v for k, v in vars(parsed_args).items() if v is not None}

    required_flags = {"output_directory"} if not help_cmd else set()
    missing = required_flags - flags.keys()
    if missing:
        raise RuntimeError(f"Missing required flags: {missing}")

    # Read video_uid_file if it is specified and put the UIDs in the "video_uids" key
    if "video_uid_file" in flags:
        if "video_uids" in flags:
            raise RuntimeError(
                "argument --video_uid_file: not allowed with argument" "--video_uids"
            )

        uids_str = Path(flags.pop("video_uid_file")).expanduser().read_text()
        flags["video_uids"] = uids_str.split()

    config = Config(**flags)

    # We need to check the universities values here since they might have been set
    # through the JSON config file, in which case the argparse checks won't validate
    # them
    universities = set(UNIV_TO_BUCKET.keys())
    unrecognized = set(config.universities) - universities
    if unrecognized:
        raise RuntimeError(
            f"Unrecognized universities: {unrecognized}. Please choose "
            f"from: {universities}"
        )

    return config



import hydra
from ego4d.features.config import FeatureExtractConfig
from ego4d.research.clep.config import TrainConfig
from ego4d.research.clep.preprocess.cc import preprocess_cc
from ego4d.research.clep.preprocess.charades import preprocess_ego_charade
from ego4d.research.clep.preprocess.ego4d_data import (
    preprocess_ego_features,
    preprocess_ego_narrations,
)
from ego4d.research.clep.preprocess.kinetics import preprocess_k400_data


@hydra.main(config_path="configs", config_name=None)
def preprocess(config: TrainConfig):
    if config.pre_config.mode == "ego4d_narr":
        preprocess_ego_narrations(config, config.pre_config.ego4d_narr)
    elif config.pre_config.mode == "ego4d_features":
        preprocess_ego_features(
            config.input_config.feature_path,
            config,
            config.pre_config.ego4d_features,
        )
    elif config.pre_config.mode == "k400":
        preprocess_k400_data(config, config.pre_config.k400)
    elif config.pre_config.mode == "ego_charade":
        preprocess_ego_charade(config, config.pre_config.ego_charade)
    elif config.pre_config.mode == "cc":
        preprocess_cc(config, config.pre_config.cc)
    else:
        raise AssertionError(f"{config.pre_config.mode} not supported")


if __name__ == "__main__":
    preprocess()  # pyre-ignore



from typing import List


def get_paths_for_commentary_time(comm: dict, t_sec: float) -> List[dict]:
    t = t_sec * 1000  # convert to ms

    paths = []
    comm_start_t = comm["start_global_time"]
    for event in comm["events"]:
        assert event["type"] == "path"
        event_t_rel = event["global_time"] - comm_start_t
        for path in event["paths"]:
            if event["action"] == "clear" and event_t_rel < t:
                # NOTE: this could be implemented more efficiently
                paths = []  # clear out the paths

            path_t_rel = path["to"]["t"] - comm["start_global_time"]
            if path_t_rel > t:
                break
            else:
                paths.append(
                    {
                        "from": {
                            "x": path["from"]["x"],
                            "y": path["from"]["y"],
                            "t": (path["from"]["t"] - event["global_time"]) / 1000.0,
                        },
                        "to": {
                            "x": path["to"]["x"],
                            "y": path["to"]["y"],
                            "t": (path["to"]["t"] - event["global_time"]) / 1000.0,
                        },
                    }
                )
    return paths



# Copyright (c) Meta Platforms, Inc. and affiliates. All Rights Reserved.

from dataclasses import dataclass
from typing import Optional

import torch
from ego4d.features.config import BaseModelConfig, InferenceConfig
from torch.nn import Module
from torchaudio.transforms import MelSpectrogram
from torchvision.transforms import Compose


@dataclass
class ModelConfig(BaseModelConfig):
    n_fft: int = 1024
    win_length: Optional[int] = None
    hop_length: int = 160  # 10ms
    n_mels: int = 128


class MelSpectrogramModel(Module):
    def __init__(self, inference_config: InferenceConfig, model_config: ModelConfig):
        super().__init__()
        self.inference_config = inference_config
        self.config = model_config

    def get_mel_spectrogram_transform(self, freq):
        return MelSpectrogram(
            sample_rate=self.inference_config.norm_config.resample_audio_rate,
            n_fft=self.config.n_fft,
            win_length=self.config.win_length,
            hop_length=self.config.hop_length,
            n_mels=self.config.n_mels,
        )

    def forward(self, x) -> torch.Tensor:
        assert len(x["audio_sample_rate"]) == 1
        f = self.get_mel_spectrogram_transform(x["audio_sample_rate"][0])
        if "audio" not in x:
            return torch.empty(1, 1)
        return f(x["audio"])


def load_model(
    inference_config: InferenceConfig,
    config: ModelConfig,
    patch_final_layer: bool = True,
) -> Module:
    model = MelSpectrogramModel(inference_config, config)
    # don't need to set to GPU for a Mel spectrogram - but will do anyway
    model = model.eval()
    model = model.to(inference_config.device)
    return model


def get_transform(inference_config: InferenceConfig, config: ModelConfig):
    # do nothing
    return Compose([])



# pyre-strict
import functools
import itertools
import os
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass

from typing import Any, Dict, List, Optional, Tuple

import pandas as pd
from ego4d.cli.manifest import VideoMetadata

from ego4d.cli.universities import BUCKET_TO_UNIV
from ego4d.internal.s3 import StreamPathMgr
from ego4d.internal.validation.ffmpeg_utils import get_video_info, VideoInfo
from ego4d.internal.validation.manifest import (
    Annotations,
    AuxiliaryVideoComponentDataFile,
    CaptureMetadataEgoExo,
    ComponentType,
    Error,
    ErrorLevel,
    load_egoexo_manifest,
    load_manifest,
    load_released_video_files,
    load_standard_metadata_files,
    load_standard_metadata_files_egoexo,
    Manifest,
    ManifestEgoExo,
    Particpant,
    StandardMetadata,
    StandardMetadataEgoExo,
    SynchronizedVideos,
    VideoComponentFile,
)

from iopath.common.file_io import PathManager
from iopath.common.s3 import S3PathHandler
from tqdm import tqdm

stream_path_mgr: Optional[StreamPathMgr] = None  # for streaming video
pathmgr = PathManager()  # for downloading files
pathmgr.register_handler(S3PathHandler(profile="default"))


@dataclass
class ReferencedFile:
    source_id: str
    source_location: str
    root_dir: str
    relative_path: str


def _validate_vcs(
    video_id_component_pair: Tuple[str, List[VideoComponentFile]],
    videos: Dict[str, VideoMetadata],
    university: str,
) -> List[Error]:
    errors = []
    video_id, components = video_id_component_pair
    components.sort(key=lambda x: x.component_index)

    if video_id not in videos:
        errors.append(
            Error(
                ErrorLevel.ERROR,
                video_id,
                "video_not_found_in_video_metadata",
                f"{video_id} in video_components_dict can't be found in video_metadata",
            )
        )
        return errors
    elif videos[video_id].number_video_components != len(components):
        errors.append(
            Error(
                ErrorLevel.ERROR,
                video_id,
                "video_component_length_inconsistent",
                f"the video has {len(components)} components when it"
                f" should have {videos[video_id].number_video_components}",
            )
        )
    components.sort(key=lambda x: x.component_index)
    # check component_index is incremental and starts at 0
    university_video_folder_path = videos[video_id].university_video_folder_path
    for i in range(len(components)):
        if components[i].component_index != i:
            errors.append(
                Error(
                    ErrorLevel.ERROR,
                    video_id,
                    "video_component_wrong_index",
                    f"the video component has index {components[i].component_index}"
                    f" when it should have {i}",
                )
            )
        if (
            not components[i].video_component_relative_path
            and not components[i].is_redacted
        ):
            errors.append(
                Error(
                    ErrorLevel.ERROR,
                    video_id,
                    "empty_video_component_relative_path",
                    f"Found an empty relative path for the video_id {video_id}",
                )
            )
        else:
            if components[i].is_redacted:
                path = os.path.join(
                    university_video_folder_path,
                    components[i].video_component_relative_path,
                )
                if not path.startswith("s3://"):
                    errors.append(
                        Error(
                            ErrorLevel.WARN,
                            video_id,
                            "using_local_path",
                            f"{path} ",
                        )
                    )

                if "s3://" in path:
                    bucket = path.split("s3://")[1].split("/")[0]
                    assoc_uni = BUCKET_TO_UNIV[bucket]
                    if assoc_uni != university:
                        errors.append(
                            Error(
                                ErrorLevel.ERROR,
                                video_id,
                                "incorrect_bucket_name",
                                path,
                            )
                        )

                if not pathmgr.exists(path):
                    errors.append(
                        Error(
                            ErrorLevel.ERROR,
                            video_id,
                            "path_does_not_exist",
                            path,
                        )
                    )

    return errors


def validate_video_components(
    manifest: Manifest,
    university: str,
    num_workers: int,
) -> List[Error]:
    """
    This method in parallel (w.r.t num_workers) checks each video component file
    provided in the manifest.

    Refer to _validate_vcs for logic
    """
    map_fn = functools.partial(
        _validate_vcs,
        videos=manifest.videos,
        university=university,
    )
    errors = []
    with ThreadPoolExecutor(num_workers) as pool:
        vals_to_map = list(manifest.video_components.items())
        for errs in tqdm(pool.map(map_fn, vals_to_map), total=len(vals_to_map)):
            errors.extend(errs)

    return errors


def validate_mp4s(video_infos: Dict[Any, List[VideoInfo]]) -> List[Error]:  # noqa
    """
    This function checks a set of MP4 files on S3 for the following properties:
    - null fps
    - consistent width/height
    - rotation consistent
    - video time base consistent
    - video codec consistent
    - audio codec consistent
    - mp4 duration too large or small
    - mp4 duration does not exist
    - video fps consistent
    - SAR consistent
    - FPS consistent
    """
    errors = []
    for video_id, vis in video_infos.items():
        total_vcodec = set()
        total_acodec = set()
        total_rotate = set()
        total_size = set()
        total_vtb = set()
        total_sar = set()
        total_fps = set()

        for i, video_info in enumerate(vis):
            if video_info is None:
                errors.append(
                    Error(
                        ErrorLevel.ERROR,
                        video_id,
                        "no_video_info_for_component",
                        f"component {i} has no metadata",
                    )
                )
                continue

            # check consistent video codec
            if video_info.vcodec is not None:
                total_vcodec.add(video_info.vcodec)

            # check consistent audio time base
            if video_info.acodec is not None:
                total_acodec.add(video_info.acodec)

            # check consistent rotation
            total_rotate.add(video_info.rotate)

            # check consistent width and height
            if (
                video_info.sample_width is not None
                and video_info.sample_height is not None
            ):
                total_size.add((video_info.sample_width, video_info.sample_height))

                if (
                    video_info.sample_width < video_info.sample_height
                    and video_info.rotate is None
                ):
                    errors.append(
                        Error(
                            ErrorLevel.ERROR,
                            video_id,
                            "component_having_width_lt_height",
                            f"component {i} has width < height without rotation",
                        )
                    )

            # check consistent video time base
            if video_info.video_time_base is not None:
                total_vtb.add(video_info.video_time_base)

            # check consistent sar
            if video_info.sar is not None:
                total_sar.add(video_info.sar)

            # check null/inconsistent video fps
            if video_info.fps is None:
                errors.append(
                    Error(
                        ErrorLevel.WARN,
                        video_id,
                        "missing_fps_info",
                        f"component {i} has null fps value",
                    )
                )
            else:
                total_fps.add(video_info.fps)

            # check null mp4 duration
            if video_info.mp4_duration is None:
                errors.append(
                    Error(
                        ErrorLevel.WARN,
                        video_id,
                        "missing_mp4_duration_info",
                        f"component {i} has no mp4 duration",
                    )
                )
            else:
                video_length = (
                    video_info.vstart + video_info.vduration
                    if video_info.vstart is not None
                    else video_info.vduration
                )
                audio_length = (
                    video_info.astart + video_info.aduration
                    if video_info.astart is not None
                    else video_info.aduration
                )
                if video_length is None and audio_length is None:
                    errors.append(
                        Error(
                            ErrorLevel.WARN,
                            video_id,
                            "no_video_or_audio_stream_duration",
                            f"component {i} has no video or audio stream duration metadata",
                        )
                    )

                if video_length is None:
                    video_length = -1

                if audio_length is None:
                    audio_length = -1

                video_stream_length = max(video_length, audio_length)
                threshold = 2 / (video_info.fps or 30)
                delta = abs(video_stream_length - video_info.mp4_duration)
                if delta >= threshold:
                    errors.append(
                        Error(
                            ErrorLevel.WARN,
                            video_id,
                            "mp4_duration_too_large_or_small",
                            f"component {i}: mp4_duration={video_info.mp4_duration}, stream_duration={video_stream_length}, vsd={video_length}, asd={audio_length}",  # noqa
                        )
                    )

        if len(total_vcodec) > 1:
            errors.append(
                Error(
                    ErrorLevel.ERROR,
                    video_id,
                    "inconsistent_video_codec",
                    "",
                )
            )

        if len(total_acodec) > 1:
            errors.append(
                Error(
                    ErrorLevel.WARN,
                    video_id,
                    "inconsistent_audio_codec",
                    "",
                )
            )

        if len(total_rotate) > 1:
            errors.append(
                Error(ErrorLevel.ERROR, video_id, "inconsistent_rotation", "")
            )

        if len(total_size) > 1:
            errors.append(
                Error(
                    ErrorLevel.ERROR,
                    video_id,
                    "inconsistent_width_height_pair",
                    "components with inconsistent width x height",
                )
            )

        if len(total_vtb) > 1:
            errors.append(
                Error(
                    ErrorLevel.ERROR,
                    video_id,
                    "inconsistent_video_time_base",
                    "",
                )
            )

        if len(total_sar) > 1:
            errors.append(Error(ErrorLevel.WARN, video_id, "inconsistent_sar", ""))

        if len(total_fps) > 1:
            errors.append(
                Error(ErrorLevel.WARN, video_id, "inconsistent_video_fps", "")
            )
    return errors


def validate_synchronized_metadata(
    video_metadata_dict: Dict[str, VideoMetadata],
    synchronized_video_dict: Dict[str, SynchronizedVideos],
) -> List[Error]:
    """
    Args:
        synchronized_video_dict: Dict[str, SynchronizedVideos]: mapping from
        video_grouping_id  to a list of all SynchronizedVideos objects that
        have equal values at video_grouping_id

        error_message: List[Error]: a list to store Error objects
        generated when validating synchronized_videos.csv.
    """
    errors = []
    if synchronized_video_dict:
        for video_grouping_id, component in synchronized_video_dict.items():
            for video_id, _ in component.associated_videos.items():
                if video_id not in video_metadata_dict:
                    errors.append(
                        Error(
                            ErrorLevel.ERROR,
                            video_id,
                            "video_not_found_in_video_metadata",
                            f"({video_id}, {video_grouping_id}) in synchronized_video_dict can't be found in video_metadata",  # noqa
                        )
                    )
    return errors


def validate_auxilliary_videos(
    video_metadata_dict: Dict[str, VideoMetadata],
    video_components_dict: Dict[str, List[VideoComponentFile]],
    auxiliary_video_component_dict: Dict[str, List[AuxiliaryVideoComponentDataFile]],
    component_types: Dict[str, ComponentType],
) -> List[Error]:
    """
    Args:
        auxiliary_video_component_dict: Dict[str, List[AuxiliaryVideoComponentDataFile]]:
        mapping from university_video_id to a list of all AuxiliaryVideoComponentDataFile
        objects that have equal values at university_video_id

        error_message: List[Error]: a list to store Error objects
        generated when validating auxilliary_videos.csv.
    """
    # Check ids in auxiliary_video_component_dict are in video_metadata_dict
    # and that the component_type is valid
    errors = []
    if auxiliary_video_component_dict:
        for video_id, aux_components in auxiliary_video_component_dict.items():
            if video_id not in video_metadata_dict:
                errors.append(
                    Error(
                        ErrorLevel.ERROR,
                        video_id,
                        "video_not_found_in_video_metadata",
                        f"{video_id} in auxiliary_video_component_dict can't be found in video_metadata",  # noqa
                    )
                )

            else:
                vcs = [
                    component.component_index
                    for component in video_components_dict[video_id]
                ]

                for component_type_id, xs in itertools.groupby(
                    aux_components, lambda x: x.component_type_id
                ):
                    xs = list(xs)
                    if video_metadata_dict[video_id].number_video_components != len(xs):
                        errors.append(
                            Error(
                                ErrorLevel.ERROR,
                                video_id,
                                "video_component_length_inconsistent",
                                f"the video has {len(aux_components)} auxiliary components when it"  # noqa
                                f" should have {video_metadata_dict[video_id].number_video_components}",  # noqa
                            )
                        )
                        continue

                    xs.sort(key=lambda x: x.component_index)  # pyre-ignore
                    for i, component in enumerate(xs):
                        if vcs[i] != component.component_index:
                            errors.append(
                                Error(
                                    ErrorLevel.ERROR,
                                    video_id,
                                    "video_component_wrong_index",
                                    "the video component has auxiliary component index",
                                    f"{component.component_index} (type_id={component_type_id})",  # noqa
                                    f" when it should have {vcs[i]}",  # noqa
                                )
                            )
                        if component.component_type_id not in component_types:
                            errors.append(
                                Error(
                                    ErrorLevel.ERROR,
                                    video_id,
                                    "component_type_id_not_found",
                                    f"auxiliary component's (type_id={component_type_id})",
                                    f"'{component.component_type_id}",
                                    "does not exist in component_types'",  # noqa
                                )
                            )
    return errors


def validate_participant(
    video_metadata_dict: Dict[str, VideoMetadata],
    participant_dict: Dict[str, List[Particpant]],
) -> List[Error]:
    """
    Args:
        participant_dict: Dict[str, Particpant]: mapping from participant_id
        to Participant objects that have equal values at participant_id

        error_message: List[Error]: a list to store Error objects
        generated when validating participants.csv.
    """
    errors = []
    if participant_dict:
        ps_in_video_metadata = {
            x.recording_participant_id for x in video_metadata_dict.values()
        }
        ps_in_meta = set(participant_dict.keys())

        wrong_ps = ps_in_video_metadata - ps_in_meta

        if len(wrong_ps) > 0:
            for participant_id in wrong_ps:
                errors.append(
                    Error(
                        ErrorLevel.ERROR,
                        participant_id,
                        "participant_metadata_not_found",
                        f"participant '{participant_id}' does not exist in participant metadata",  # noqa
                    )
                )

    else:
        errors.append(Error(ErrorLevel.WARN, "general", "no_participant_metadata", ""))
    return errors


def validate_annotations(
    video_metadata_dict: Dict[str, VideoMetadata],
    annotations_dict: Dict[str, Annotations],
) -> List[Error]:
    """
    Args:
        annotations_dict: Dict[str, Annotations]: mapping from participant_id
        to Participant objects that have equal values at participant_id

        error_message: List[Error]: a list to store Error objects
        generated when validating participants.csv.
    """
    print("validating annotations")
    errors = []
    if annotations_dict:
        for video_id in annotations_dict:
            if video_id not in video_metadata_dict:
                errors.append(
                    Error(
                        ErrorLevel.ERROR,
                        video_id,
                        "video_not_found_in_video_metadata",
                        f"{video_id} in annotations_dict can't be found in video_metadata",
                    )
                )
    return errors


def validate_video_metadata(
    manifest: Manifest,
    standard_metadata: StandardMetadata,
) -> List[Error]:
    # Check from video_metadata:
    errors = []
    video_ids = set()
    for video_id, video_metadata in manifest.videos.items():
        # 0. check no duplicate university_video_id
        if video_id in video_ids:
            errors.append(
                Error(
                    ErrorLevel.ERROR,
                    video_id,
                    "duplicate_video_id",
                    "duplicate video id in metadata",
                )
            )
        video_ids.add(video_id)
        # 1. participant_ids are in participant_dict
        if manifest.participants is not None:
            if video_metadata.recording_participant_id is None:
                errors.append(
                    Error(
                        ErrorLevel.ERROR,
                        video_id,
                        "null_participant_id",
                        "",
                    )
                )

        # 2. scenario_ids are in scenarios
        for scenario_id in standard_metadata.scenarios:
            if scenario_id not in standard_metadata.scenarios:
                errors.append(
                    Error(
                        ErrorLevel.ERROR,
                        video_id,
                        "scenario_id_not_found",
                        f"video_scenario_id: '{scenario_id}' not in scenarios.csv",
                    )
                )

        # 3. device_ids are in devices
        if video_metadata.device_id is None:
            errors.append(Error(ErrorLevel.WARN, video_id, "device_id_null", ""))
        elif video_metadata.device_id not in standard_metadata.devices:
            errors.append(
                Error(
                    ErrorLevel.ERROR,
                    video_id,
                    "device_id_not_found",
                    f"device_id '{video_metadata.device_id}' not in devices.csv",
                )
            )

        # 4. physical_settings are in physical_setting
        if (
            video_metadata.physical_setting_id
            and video_metadata.physical_setting_id not in manifest.physical_setting
        ):
            errors.append(
                Error(
                    ErrorLevel.ERROR,
                    video_id,
                    "physical_setting_id_not_found",
                    f"physical_setting_id '{video_metadata.physical_setting_id}' not in physical_setting.csv",  # noqa
                )
            )

        # 5. university_video_ids are in components
        if video_id not in manifest.video_components:
            errors.append(
                Error(
                    ErrorLevel.ERROR,
                    video_id,
                    "video_not_found_in_video_components",
                    f"{video_id} in video_metadata can't be found in video_components_dict",
                )
            )
    return errors


def _get_video_metadata_map_fn(
    uid_path_pair: Tuple[str, str],
) -> Tuple[str, Optional[VideoInfo], Optional[Error]]:
    assert stream_path_mgr is not None
    uid, path = uid_path_pair
    metadata, errs = get_video_info(path, name=uid)
    return uid, metadata, errs


def get_video_metadata(
    video_paths: Dict[Any, str],
    num_workers: int,
) -> Tuple[Dict[str, VideoMetadata], List[Error]]:
    errors = []
    metadata = {}

    # prefetch the presigned URLs
    paths_to_check = {}
    for k, path in video_paths.items():
        paths_to_check[k] = stream_path_mgr.open(path)

    iterables = list(paths_to_check.items())
    with ThreadPoolExecutor(max_workers=num_workers) as pool:
        for uid, m, errs in tqdm(
            pool.map(_get_video_metadata_map_fn, iterables), total=len(iterables)
        ):
            metadata[uid] = m
            if errs is not None:
                errors.append(errs)
    return metadata, errors


def validate_ego4d_files(  # noqa :C901
    university: str,
    manifest: Manifest,
    metadata: StandardMetadata,
    num_workers: int,
) -> List[Error]:
    errors = []
    errors.extend(
        validate_video_components(
            manifest=manifest,
            university=university,
            num_workers=num_workers,
        )
    )

    vps = {
        (vc.university_video_id, vc.component_index): os.path.join(
            manifest.videos[video_id].university_video_folder_path,
            vc.video_component_relative_path,
        )
        for video_id, vcs in manifest.video_components.items()
        for vc in vcs
    }
    print("Obtaining MP4 metadata")
    video_infos, video_errs = get_video_metadata(vps, num_workers)
    errors.extend(video_errs)

    video_infos_by_video_id = defaultdict(list)
    for (v_uid, comp_idx), vi in video_infos.items():
        video_infos_by_video_id[v_uid].append((comp_idx, vi))

    for vs in video_infos_by_video_id.values():
        vs.sort(key=lambda x: x[0])

    print("Validating MP4 files")
    errors.extend(
        validate_mp4s(
            {k: [x[1] for x in vs] for k, vs in video_infos_by_video_id.items()}
        )
    )
    print("Validating synchronized video metadata")
    errors.extend(
        validate_synchronized_metadata(
            video_metadata_dict=manifest.videos,
            synchronized_video_dict=manifest.sync_videos,
        )
    )
    print("Validating auxiliary video metadata")
    errors.extend(
        validate_auxilliary_videos(
            video_metadata_dict=manifest.videos,
            video_components_dict=manifest.video_components,
            auxiliary_video_component_dict=manifest.aux_components,
            component_types=metadata.component_types,
        )
    )
    print("Validating participant metadata")
    errors.extend(
        validate_participant(
            manifest.videos,
            manifest.participants,
        )
    )
    print("validating annotations")
    errors.extend(
        validate_annotations(
            manifest.videos,
            manifest.annotations,
        )
    )
    print("validating video metadata")
    errors.extend(
        validate_video_metadata(
            manifest=manifest,
            standard_metadata=metadata,
        )
    )
    return errors


def summarize_errors(
    errors: List[Error],
    released_videos: Optional[Dict[str, List[str]]],
    university: str,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    error_dict_non_released = defaultdict(int)
    error_dict_released = defaultdict(int)
    if len(errors) > 0 and released_videos is not None:
        for err in errors:
            if err.uid not in released_videos[university]:
                error_dict_non_released[err.errorType] += 1
            else:
                error_dict_released[err.errorType] += 1

    errors_dict = {
        "error_class": [],
        "uid": [],
        "error_type": [],
        "description": [],
        "is_released": [],
    }

    err_by_type_in_release = defaultdict(int)
    err_by_type_total = defaultdict(int)
    for e in errors:
        clazz = "error" if e.level == ErrorLevel.ERROR else "warning"
        errors_dict["error_class"].append(clazz)
        errors_dict["uid"].append(e.uid)
        errors_dict["error_type"].append(e.type)
        errors_dict["description"].append(e.description)
        errors_dict["is_released"].append(
            e.uid in released_videos if released_videos is not None else None
        )

        if released_videos is not None and e.uid in released_videos:
            err_by_type_in_release[(clazz, e.type)] += 1
        err_by_type_total[(clazz, e.type)] += 1

    summary_dict = {
        "error_class": [],
        "error_type": [],
        "num_total": [],
        "num_in_release": [],
    }

    for (clazz, et), t in err_by_type_total.items():
        summary_dict["error_class"].append(clazz)
        summary_dict["error_type"].append(et)
        summary_dict["num_total"].append(t)
        summary_dict["num_in_release"].append(err_by_type_in_release.get(et, None))

    errors_df = pd.DataFrame(errors_dict)
    summary_df = pd.DataFrame(summary_dict)
    return errors_df, summary_df


def _check_associated_takes_metadata(
    manifest: ManifestEgoExo,
    metadata: StandardMetadataEgoExo,
    capture: CaptureMetadataEgoExo,
    capture_uid: str,
) -> List[Error]:
    ret = []
    if capture_uid not in manifest.takes:
        ret.append(
            Error(
                ErrorLevel.ERROR,
                capture_uid,
                "no_takes_for_capture",
                f"capture {capture_uid} has no associated takes",
            )
        )
        return ret

    takes = manifest.takes[capture_uid]
    if len(takes) != capture.number_takes:
        ret.append(
            Error(
                ErrorLevel.ERROR,
                capture_uid,
                "incorrect_number_takes_for_capture",
                f"capture {capture_uid} has {capture.number_takes} specified but only recieved {len(takes)} takes",
            )
        )

    seen_take_ids = set()
    for take in takes:
        if take.take_id in seen_take_ids:
            ret.append(
                Error(
                    ErrorLevel.ERROR,
                    take.take_id,
                    "repeated_take_id",
                    f"take {take.take_id} for capture {capture_uid} is listed multiple times",
                )
            )
        seen_take_ids.add(take.take_id)
        if (
            take.scenario_id is None or take.scenario_id not in metadata.scenarios
        ) and not take.is_dropped:
            ret.append(
                Error(
                    ErrorLevel.ERROR,
                    take.take_id,
                    "take_has_incorrect_scenario_id",
                    f"take {take.take_id} for capture {capture_uid} has incorrect scenario ID {take.scenario_id}",
                )
            )
        if take.recording_participant_id is None and not take.is_dropped:
            ret.append(
                Error(
                    ErrorLevel.WARN,
                    take.take_id,
                    "take_missing_participant",
                    f"take {take.take_id} for capture {capture_uid} has missing participant ID",
                )
            )

        if take.object_ids is None:
            ret.append(
                Error(
                    ErrorLevel.WARN,
                    take.take_id,
                    "take_missing_objects",
                    f"take {take.take_id} has no objects provided. Provide an empty list to silence warning",
                )
            )

        additional_objects = set(take.object_ids or {}) - set(manifest.objects or {})
        if len(additional_objects) > 0:
            ret.append(
                Error(
                    ErrorLevel.ERROR,
                    take.take_id,
                    "take_has_unspecified_objects",
                    f"take {take.take_id} for capture {capture_uid} has unspecified objects: {additional_objects}",
                )
            )

    return ret


def _check_capture_metadata(
    manifest: ManifestEgoExo,
    metadata: StandardMetadataEgoExo,
    capture: CaptureMetadataEgoExo,
    capture_uid: str,
) -> List[Error]:
    ret = []
    capture = manifest.captures[capture_uid]
    if capture.start_date_recorded_utc is None:
        ret.append(
            Error(
                ErrorLevel.WARN,
                capture_uid,
                "capture_missing_recording_date",
                f"capture {capture_uid} has no recording date",
            )
        )

    return ret


def _check_objects(manifest: ManifestEgoExo) -> List[Error]:
    ret = []
    physical_setting_ids = (
        set(manifest.physical_setting) if manifest.physical_setting is not None else {}
    )
    assert manifest.objects is not None
    for object_id, obj in manifest.objects.items():
        if obj.physical_setting_id not in physical_setting_ids:
            ret.append(
                Error(
                    ErrorLevel.ERROR,
                    object_id,
                    "object_links_to_invalid_physical_setting_id",
                    f"phyiscal setting {obj.physical_setting_id} does not exist",
                )
            )
    return ret


def _check_participants(
    manifest: ManifestEgoExo, metadata: StandardMetadataEgoExo
) -> List[Error]:
    ret = []
    assert manifest.participants is not None
    for (participant_id, scenario_id, _), p in manifest.participants.items():
        if scenario_id is not None and scenario_id not in metadata.scenarios:
            ret.append(
                Error(
                    ErrorLevel.ERROR,
                    participant_id,
                    "participant_links_to_invalid_scenario",
                    f"scenario {p.scenario_id} does not exist",
                )
            )

        if len(p.pre_survey_data) > 0 and len(p.participant_metadata) > 0:
            ret.append(
                Error(
                    ErrorLevel.ERROR,
                    participant_id,
                    "participant_has_both_pre_survey_data_and_participant_metadata",
                    "in a single entry, participants should have either pre survey data or participant_metadata populated. not both.",
                )
            )
        elif len(p.pre_survey_data) == 0 and len(p.participant_metadata) == 0:
            ret.append(
                Error(
                    ErrorLevel.ERROR,
                    participant_id,
                    "participant_has_empty_pre_survey_data_and_participant_metadata",
                    "participants should have either pre survey data or participant_metadata populated",
                )
            )

        else:
            if len(p.pre_survey_data) > 0:
                if p.scenario_id is None or p.scenario_id == 0:
                    ret.append(
                        Error(
                            ErrorLevel.ERROR,
                            participant_id,
                            "participant_has_pre_survey_data_without_scenario_id",
                            "If pre_survey_data is provided, then a valid scenario_id must be provided",
                        )
                    )

                ks = {
                    "recording_location",
                    "scenario_num_iterations",
                    "scenario_frequency",
                    "scenario_experience_years",
                    "has_taught_scenario",
                    "has_recorded_scenario_howto",
                    "has_watched_others_scenario_videos",
                    "has_qualifications_or_professional_training",
                    "typical_time_to_complete_scenario_minutes",
                    "typical_time_per_practice_session_minutes",
                }
                given_ks = set(p.pre_survey_data.keys())

                missing_ks = ks - given_ks
                aux_ks = given_ks - ks
                if len(missing_ks) > 0 or len(aux_ks) > 0:
                    ret.append(
                        Error(
                            ErrorLevel.WARN,
                            participant_id,
                            "participant_pre_survey_data_contraints",
                            f"auxiliary keys: {aux_ks}, missing keys: {missing_ks}",
                        )
                    )

                if "recording_location" in given_ks:
                    v = p.pre_survey_data["recording_location"]
                    valid_vs = {
                        "typical",
                        "familiar",
                        "unfamiliar",
                    }
                    if v not in valid_vs:
                        ret.append(
                            Error(
                                ErrorLevel.WARN,
                                participant_id,
                                "participant_pre_survey_data_contraints",
                                f"recording_location scenario_num_iterations {v} not one of: {valid_vs}",
                            )
                        )

                if "scenario_num_iterations" in given_ks:
                    v = p.pre_survey_data["scenario_num_iterations"]
                    valid_vs = {
                        "1-10",
                        "10-100",
                        "100-500",
                        "500-1000",
                        "1000+",
                    }
                    if v not in valid_vs:
                        ret.append(
                            Error(
                                ErrorLevel.WARN,
                                participant_id,
                                "participant_pre_survey_data_contraints",
                                f"scenario_num_iterations {v} not one of: {valid_vs}",
                            )
                        )

                if "scenario_frequency" in given_ks:
                    v = p.pre_survey_data["scenario_frequency"]
                    valid_vs = {
                        "daily",
                        "weekly",
                        "monthly",
                        "rarely",
                        "never",
                    }
                    if v not in valid_vs:
                        ret.append(
                            Error(
                                ErrorLevel.WARN,
                                participant_id,
                                "participant_pre_survey_data_contraints",
                                f"scenario_frequency {v} not one of: {valid_vs}",
                            )
                        )

                if "scenario_experience_years" in given_ks:
                    v = p.pre_survey_data["scenario_experience_years"]
                    valid_vs = {
                        "1 year",
                        "1-3 years",
                        "3-5 years",
                        "5-10 years",
                        "10+ years",
                    }
                    if v not in valid_vs:
                        ret.append(
                            Error(
                                ErrorLevel.WARN,
                                participant_id,
                                "participant_pre_survey_data_contraints",
                                f"scenario_experience_years {v} not one of: {valid_vs}",
                            )
                        )
                if "has_taught_scenario" in given_ks:
                    v = p.pre_survey_data["has_taught_scenario"]
                    try:
                        _ = bool(v)
                    except Exception as e:
                        ret.append(
                            Error(
                                ErrorLevel.WARN,
                                participant_id,
                                "participant_pre_survey_data_contraints",
                                f"has_taught_scenario could not be converted to boolean: {e}",
                            )
                        )
                if "has_recorded_scenario_howto" in given_ks:
                    v = p.pre_survey_data["has_recorded_scenario_howto"]
                    try:
                        _ = bool(v)
                    except Exception as e:
                        ret.append(
                            Error(
                                ErrorLevel.WARN,
                                participant_id,
                                "participant_pre_survey_data_contraints",
                                f"has_recorded_scenario_howto could not be converted to boolean: {e}",
                            )
                        )

                if "typical_time_to_complete_scenario_minutes" in given_ks:
                    v = p.pre_survey_data["typical_time_to_complete_scenario_minutes"]
                    try:
                        _ = int(v)
                    except Exception as e:
                        ret.append(
                            Error(
                                ErrorLevel.WARN,
                                participant_id,
                                "participant_pre_survey_data_constraints",
                                f"typical_time_to_complete_scenario_minutes could not be converted to integer: {e}",
                            )
                        )

            if len(p.participant_metadata) > 0:
                ks = {
                    "age_range",
                    "gender",
                    "race_ethnicity",
                    "country_born",
                    "native_language",
                    "home_language",
                    "education_completed",
                    "current_student",
                    "field",
                }
                given_ks = set(p.participant_metadata.keys())
                missing_ks = ks - given_ks
                aux_ks = given_ks - ks
                if len(missing_ks) > 0 or len(aux_ks) > 0:
                    ret.append(
                        Error(
                            ErrorLevel.WARN,
                            participant_id,
                            "participant_metadata_constraints",
                            f"auxiliary keys: {aux_ks}, missing keys: {missing_ks}",
                        )
                    )

                if "gender" in p.participant_metadata:
                    v = p.participant_metadata["gender"]
                    valid_vs = {
                        "female",
                        "male",
                        "non_binary",
                        "prefer_not_to_disclose",
                        "prefer_to_self_describe",
                    }
                    if v not in valid_vs:
                        ret.append(
                            Error(
                                ErrorLevel.WARN,
                                participant_id,
                                "participant_metadata_constraints",
                                f"gender '{v}' not one of: {valid_vs}",
                            )
                        )

    return ret


def _check_video_metadata(
    manifest: ManifestEgoExo, metadata: StandardMetadataEgoExo
) -> List[Error]:
    ret = []
    for capture_uid, videos in manifest.videos.items():
        any_has_walkaround = any(v.has_walkaround for v in videos)
        any_is_ego = any(v.is_ego for v in videos)
        if not any_has_walkaround:
            ret.append(
                Error(
                    ErrorLevel.ERROR,
                    capture_uid,
                    "capture_missing_walkaround",
                    "capture does not have walkaround (from video metadata)",
                )
            )

        if not any_is_ego:
            ret.append(
                Error(
                    ErrorLevel.ERROR,
                    capture_uid,
                    "capture_missing_ego",
                    "capture does not have ego view",
                )
            )

        seen_devices = set()
        for video in videos:
            video_uid = (capture_uid, video.university_video_id)
            video_uid_str = (
                f"capture_uid={capture_uid},video_id={video.university_video_id}"
            )
            if capture_uid not in manifest.captures:
                ret.append(
                    Error(
                        ErrorLevel.ERROR,
                        video_uid_str,
                        "video_invalid_capture_uid_fk",
                        f"video links to invalid capture_uid {capture_uid}",
                    )
                )
            if video_uid not in manifest.video_components:
                ret.append(
                    Error(
                        ErrorLevel.ERROR,
                        video_uid_str,
                        "video_has_no_linked_components",
                        "video has no associated video components",
                    )
                )
            else:
                vcs = manifest.video_components[video_uid]  # pyre-ignore
                if len(vcs) != video.number_video_components:
                    ret.append(
                        Error(
                            ErrorLevel.ERROR,
                            video_uid_str,
                            "video_has_incorrect_num_components",
                            f"video specified {video.number_video_components} linked components but has {len(vcs)} in metadata",
                        )
                    )

                all_comps_red = all(vc.is_redacted for vc in vcs)
                if video.is_redacted and not all_comps_red:
                    ret.append(
                        Error(
                            ErrorLevel.ERROR,
                            video_uid_str,
                            "video_associated_components_not_redacted",
                            f"video flagged as redacted but components are not all redacted",
                        )
                    )

            if video.device_type not in metadata.devices:
                ret.append(
                    Error(
                        ErrorLevel.ERROR,
                        video_uid_str,
                        "video_invalid_device",
                        f"video has invalid device_type: {video.device_type}",
                    )
                )
            device_uid = (video.device_type, video.device_id)
            if device_uid in seen_devices:
                ret.append(
                    Error(
                        ErrorLevel.ERROR,
                        video_uid_str,
                        "video_duplicate_device",
                        f"video has duplicate device type and instance: device_type={video.device_type}, device_id (instance)={video.device_id}",
                    )
                )
            seen_devices.add(device_uid)
    return ret


def _check_video_components(manifest: ManifestEgoExo) -> List[Error]:
    ret = []
    all_video_ids = set(
        v.university_video_id
        for vs in manifest.videos.values()
        for v in vs
        if v.number_video_components > 0
    )
    for (capture_uid, video_id), vcs in manifest.video_components.items():
        next_expected_idx = 0
        for vc in sorted(vcs, key=lambda x: x.component_index):
            key_str = f"(capture={capture_uid}, video_id={video_id}, component_idx={vc.component_index})"
            if video_id not in all_video_ids:
                ret.append(
                    Error(
                        ErrorLevel.ERROR,
                        key_str,
                        "video_component_wrong_video_fk",
                        f"video component has incorrect video id: {video_id}",
                    )
                )
                continue

            if capture_uid not in manifest.captures:
                assert capture_uid == vc.university_capture_id
                ret.append(
                    Error(
                        ErrorLevel.ERROR,
                        key_str,
                        "video_component_wrong_capture_uid_fk",
                        f"video component has incorrect university_capture_id: {vc.university_capture_id}",
                    )
                )
                continue

            if vc.component_index != next_expected_idx:
                ret.append(
                    Error(
                        ErrorLevel.ERROR,
                        key_str,
                        "video_component_incorrect_index",
                        f"expected index: {next_expected_idx}, got {vc.component_index}",
                    )
                )
            next_expected_idx += 1
    return ret


def _get_referenced_files(manifest: ManifestEgoExo) -> List[ReferencedFile]:
    ret = []
    for capture in manifest.captures.values():
        ret.append(
            ReferencedFile(
                source_id=capture.university_capture_id,
                source_location="capture_post_survey",
                root_dir=capture.university_video_folder_path,
                relative_path=capture.post_surveys_relative_path,
            )
        )

    for vcs in manifest.video_components.values():
        for vc in vcs:
            if vc.university_capture_id not in manifest.captures:
                continue

            root_dir = manifest.captures[
                vc.university_capture_id
            ].university_video_folder_path
            ret.append(
                ReferencedFile(
                    source_id=f"(capture={vc.university_capture_id}, video_id={vc.university_video_id}, idx={vc.component_index})",
                    source_location="video_component",
                    root_dir=root_dir,
                    relative_path=vc.video_component_relative_path,
                )
            )

    if manifest.objects:
        for object_id, obj in manifest.objects.items():
            if obj.university_object_id not in manifest.captures:
                continue
            root_dir = manifest.captures[
                obj.university_object_id
            ].university_video_folder_path
            ret.append(
                ReferencedFile(
                    source_id=f"(object_id={object_id})",
                    source_location="object",
                    root_dir=root_dir,
                    relative_path=obj.object_relative_path,
                )
            )
    return ret


def _check_files_exist(
    files: List[ReferencedFile], num_workers: int, university: str
) -> List[Error]:
    ret = []

    def _maybe_strip_last_slash(path: str) -> str:
        if path.endswith("/"):
            return path[0:-1]
        return path

    def _check_file(f: ReferencedFile) -> List[Error]:
        errs = []
        assert f.root_dir is not None
        if f.relative_path is not None and f.relative_path.startswith("s3"):
            path = f.relative_path
        else:
            path = os.path.join(f.root_dir, f.relative_path or "")

        if path is None or _maybe_strip_last_slash(
            f.root_dir
        ) == _maybe_strip_last_slash(path):
            errs.append(
                Error(
                    ErrorLevel.ERROR,
                    f.source_id,
                    "path_is_empty_or_null",
                    f"source=(location={f.source_location}, id={f.source_id})",
                )
            )
            return errs

        if not pathmgr.exists(path):
            errs.append(
                Error(
                    ErrorLevel.ERROR,
                    f.source_id,
                    "path_does_not_exist",
                    f"{path}; source=(location={f.source_location}, id={f.source_id})",
                )
            )
            return errs

        if not path.startswith("s3://"):
            errs.append(
                Error(
                    ErrorLevel.ERROR,
                    f.source_id,
                    "path_not_s3_path",
                    f"{path} is not on S3 - please fix this before uploading",
                )
            )
        else:
            bucket = path.split("s3://")[1].split("/")[0]
            univ_from_path = BUCKET_TO_UNIV[bucket]
            if univ_from_path != university:
                errs.append(
                    Error(
                        ErrorLevel.ERROR,
                        f.source_id,
                        "s3_path_not_in_university_bucket",
                        f"{path} is not in {university}'s bucket ({BUCKET_TO_UNIV[university]})",
                    )
                )
        return errs

    with ThreadPoolExecutor(num_workers) as pool:
        for errs in tqdm(pool.map(_check_file, files), total=len(files)):
            ret.extend(errs)
    return ret


def _check_mp4_files(manifest: ManifestEgoExo, num_workers: int) -> List[Error]:
    ret = []
    vps = {
        (
            vc.university_capture_id,
            vc.university_video_id,
            vc.component_index,
        ): (
            os.path.join(
                manifest.captures[
                    vc.university_capture_id
                ].university_video_folder_path,
                vc.video_component_relative_path,
            )
            if not vc.video_component_relative_path.startswith("s3")
            else vc.video_component_relative_path
        )
        for vcs in manifest.video_components.values()
        for vc in vcs
        if vc.university_capture_id in manifest.captures
        and vc.video_component_relative_path.lower().endswith(".mp4")
    }

    # prefetch the presigned URLs
    print("Fetching pre-signed urls")
    assert stream_path_mgr is not None
    paths_to_check = {}
    for k, path in vps.items():
        paths_to_check[k] = stream_path_mgr.open(path)

    print(f"Obtaining metadata, using {num_workers} workers")
    metadata = {}
    iterables = list(paths_to_check.items())
    with ThreadPoolExecutor(max_workers=num_workers) as pool:
        for k, m, errs in tqdm(
            pool.map(_get_video_metadata_map_fn, iterables), total=len(iterables)
        ):
            metadata[k] = m
            if errs is not None:
                ret.append(errs)

    video_infos_by_video_id = defaultdict(list)
    for (capture_uid, v_uid, comp_idx), vi in metadata.items():
        video_infos_by_video_id[(capture_uid, v_uid)].append((comp_idx, vi))

    for vs in video_infos_by_video_id.values():
        vs.sort(key=lambda x: x[0])

    print("Validating metadata")
    ret.extend(
        validate_mp4s(
            {
                f"(capture_uid={k[0]},video_id={k[1]})": [x[1] for x in vs]
                for k, vs in video_infos_by_video_id.items()
            }
        )
    )
    return ret


def validate_egoexo_files(
    university: str,
    manifest: ManifestEgoExo,
    metadata: StandardMetadataEgoExo,
    num_workers: int,
    skip_mp4_check: bool,
) -> List[Error]:
    ret = []

    # Pre Survey Checks - Flagging if participant + scenario does not have matching pre_survey
    parts_by_key = {}
    for (participant_id, scenario_id, _), p in manifest.participants.items():

        key = (participant_id, scenario_id)
        if key not in parts_by_key:
            parts_by_key[key] = []
        parts_by_key[key].append(p)

    for takes in manifest.takes.values():
        for take in takes:
            if take.is_dropped:
                continue

            key = (take.recording_participant_id, take.scenario_id)
            if key not in parts_by_key and take.scenario_id is not None:
                higher_level_scenario_id = ((take.scenario_id) // 1000) * 1000
                new_key = (take.recording_participant_id, higher_level_scenario_id)

                if new_key in parts_by_key:  # output flag higher level task provided
                    ret.append(
                        Error(
                            ErrorLevel.ERROR,
                            take.take_id,
                            "take_no_pre_survey_higher_level_id_provided",
                            f"take {take.take_id} (participant={take.recording_participant_id}, scenario {take.scenario_id}) does not have a corresponding pre_survey. Can match with higher level id = {higher_level_scenario_id} from take's scenario ID = {take.scenario_id}",
                        )
                    )
                    key = new_key
                else:
                    ret.append(
                        Error(
                            ErrorLevel.ERROR,
                            take.take_id,
                            "take_no_pre_survey",
                            f"take {take.take_id} (participant={take.recording_participant_id}, scenario {take.scenario_id}) does not have a corresponding pre_survey.",
                        )
                    )
            elif key not in parts_by_key:
                ret.append(
                    Error(
                        ErrorLevel.ERROR,
                        take.take_id,
                        "take_no_pre_survey",
                        f"take {take.take_id} (participant={take.recording_participant_id}, scenario {take.scenario_id}) does not have a corresponding pre_survey.",
                    )
                )

            if key in parts_by_key and len(parts_by_key[key]) > 1:
                if not all(
                    x.pre_survey_data == parts_by_key[key][0].pre_survey_data
                    for x in parts_by_key[key]
                ):
                    ret.append(
                        Error(
                            ErrorLevel.ERROR,
                            take.take_id,
                            "take_multiple_pre_survey",
                            f"take {take.take_id} (participant={take.recording_participant_id}, scenario {take.scenario_id}) has multiple matching pre-survey data (len={len(parts_by_key[key])}) where at least one pre-survey data match differs.",
                        )
                    )

    # Post Survey Checks - Flagging if participant + take does not have matchin post_survey
    def get_post_survey_paths(
        captures: List[CaptureMetadataEgoExo],
    ) -> Dict[str, ReferencedFile]:
        dict_post_surveys = {}
        for capture in captures:
            if not capture.post_surveys_relative_path.startswith("s3://"):
                survey_path = os.path.join(
                    capture.university_video_folder_path,
                    capture.post_surveys_relative_path,
                )  # make absolute path if not allready
            else:
                survey_path = capture.post_surveys_relative_path
            dict_post_surveys[capture.university_capture_id] = survey_path
        return dict_post_surveys

    post_survey_files = get_post_survey_paths(list(manifest.captures.values()))

    def open_csv(capture_id_file_path):
        capture_id, file_path = capture_id_file_path
        df = None
        if pathmgr.exists(file_path):
            df = pd.read_csv(pathmgr.open(file_path), dtype=object)
        return capture_id, df

    post_by_capture_id = {}
    with ThreadPoolExecutor(num_workers) as pool:
        vals_to_map = list(post_survey_files.items())
        map_fn = open_csv
        print("Reading Post Survey Data...")
        for capture_id, df in tqdm(
            pool.map(map_fn, vals_to_map), total=len(vals_to_map)
        ):
            if df is None:
                ret.append(
                    Error(
                        ErrorLevel.ERROR,
                        capture_id,
                        "path_does_not_exist",
                        f"capture ({capture_id}) has  does not exist",
                    )
                )
                continue

            post_survey_data = df
            tid_key = [
                k
                for k in post_survey_data.columns  # noqa
                if ("Take ID" in k) or ("Take_ID" in k) or ("take_id" in k)
            ]
            post_data_path = post_survey_files[capture_id]
            if len(tid_key) != 1:
                # post_survey exists but malformed csv, no columns
                if len(post_survey_data.columns) == 0:  # noqa
                    ret.append(
                        Error(
                            ErrorLevel.ERROR,
                            capture_id,
                            "post_survey_csv_structure_no_columns",
                            f"Postsurvey data for capture {capture_id} does not provide names for each column within the CSV, file: {post_data_path}",
                        )
                    )
                else:
                    ret.append(
                        Error(
                            ErrorLevel.ERROR,
                            capture_id,
                            "post_survey_csv_structure_no_take_id_column",
                            f"Postsurvey data for capture {capture_id} does not provide a take_id column as 'take_id' in the post_survey data, file: {post_data_path}",
                        )
                    )

                post_survey_data = None
            else:
                if (
                    tid_key[0] != "take_id"
                ):  # if tid_key exists, check if it is spelled correctly
                    ret.append(
                        Error(
                            ErrorLevel.WARN,
                            capture_id,
                            "post_survey_csv_structure_take_id_column_mispelled",
                            f"'take_id' column is noted as '{tid_key[0]}' NOT 'take_id', file: {post_data_path}, capture_id={capture_id}",
                        )
                    )
            post_by_capture_id[capture_id] = df

    for takes in manifest.takes.values():
        for take in takes:
            if take.is_dropped:
                continue

            post_survey_data = post_by_capture_id.get(take.university_capture_id)

            if post_survey_data is None:
                ret.append(
                    Error(
                        ErrorLevel.ERROR,
                        take.take_id,
                        "take_no_post_survey",
                        f"take {take.take_id} does not have post survey data (capture={take.university_capture_id})",
                    )
                )
            else:
                tid_key = [
                    k
                    for k in post_survey_data.columns
                    if ("Take ID" in k) or ("Take_ID" in k) or ("take_id" in k)
                ]

                if len(tid_key) > 0:
                    tid_key = tid_key[0]
                    post_survey_data = post_survey_data[
                        post_survey_data[tid_key] == take.take_id
                    ]
                    post_survey_data = (
                        post_survey_data.iloc[0] if len(post_survey_data) > 0 else None
                    )
                else:
                    post_survey_data = None

                if post_survey_data is None:
                    ret.append(
                        Error(
                            ErrorLevel.ERROR,
                            take.take_id,
                            "take_no_post_survey_match",
                            f"take {take.take_id} does not have post survey data (capture={take.university_capture_id})",
                        )
                    )

    for capture_uid, capture in manifest.captures.items():
        ret.extend(_check_capture_metadata(manifest, metadata, capture, capture_uid))
        ret.extend(
            _check_associated_takes_metadata(manifest, metadata, capture, capture_uid)
        )

    if not manifest.participants:
        ret.append(
            Error(
                ErrorLevel.ERROR,
                "participants",
                "no_participant_metadata",
                "missing participant metadata (null or empty)",
            )
        )
    else:
        ret.extend(_check_participants(manifest, metadata))

    if not manifest.physical_setting:
        ret.append(
            Error(
                ErrorLevel.ERROR,
                "physical_setting",
                "no_physical_setting_metadata",
                "missing physical setting metadata (null or empty)",
            )
        )

    if not manifest.objects:
        ret.append(
            Error(
                ErrorLevel.ERROR,
                "objects",
                "no_objects_metadata",
                "missing object metadata (null or empty)",
            )
        )
    else:
        ret.extend(_check_objects(manifest))

    provided_capture_uids = set(manifest.captures.keys())
    provided_participant_ids = {x[0] for x in manifest.participants.keys()}
    for capture_uid, takes in manifest.takes.items():
        take = takes[0]
        if capture_uid not in provided_capture_uids:
            ret.append(
                Error(
                    ErrorLevel.ERROR,
                    take.take_id,
                    "take_has_incorrect_capture_uid",
                    f"take {take.take_id} has incorrect capture uid: {capture_uid} (typo or misisng capture in metadata?)",
                )
            )
        for take in takes:
            if take.recording_participant_id is None:
                continue
            elif take.recording_participant_id not in provided_participant_ids:
                ret.append(
                    Error(
                        ErrorLevel.ERROR,
                        take.take_id,
                        "participant_id_missing",
                        f"participant {take.recording_participant_id} is missing from participant_metadata, but referenced in capture {capture_uid} take {take.take_id}",
                    )
                )

    ret.extend(_check_video_metadata(manifest, metadata))
    ret.extend(_check_video_components(manifest))

    all_references_files = _get_referenced_files(manifest)
    print("Checking whether files exist")
    ret.extend(_check_files_exist(all_references_files, num_workers, university))
    if skip_mp4_check:
        print("Skipping MP4 file checks")
    else:
        print("Checking MP4 files")
        ret.extend(_check_mp4_files(manifest, num_workers))
    return ret


def run_validation(
    manifest_dir: str,
    standard_metadata_folder: str,
    input_university: str,
    num_workers: int,
    expiry_time_sec: int,
    released_video_path: str,
    version: str,
    output_dir: str,
    skip_mp4_check: bool,
):
    global stream_path_mgr
    if stream_path_mgr is not None:
        raise AssertionError(
            "Don't use this method in multi-threaded contexts. Use processes instead."
        )

    stream_path_mgr = StreamPathMgr(expiration_sec=expiry_time_sec)

    # get access to metadata_folder
    if version.lower() == "egoexo":
        metadata = load_standard_metadata_files_egoexo(standard_metadata_folder)
        manifest = load_egoexo_manifest(manifest_dir)
        released_videos = None  # TODO: load ingested videos here

        errors = validate_egoexo_files(
            university=input_university,
            manifest=manifest,
            metadata=metadata,
            num_workers=num_workers,
            skip_mp4_check=skip_mp4_check,
        )
    else:
        assert version.lower() == "ego4d", "expected ego4d as version"
        metadata = load_standard_metadata_files(standard_metadata_folder)
        manifest = load_manifest(manifest_dir)
        released_videos = load_released_video_files(released_video_path)

        errors = validate_ego4d_files(
            university=input_university,
            manifest=manifest,
            metadata=metadata,
            num_workers=num_workers,
        )

    errors_df, summary_df = summarize_errors(
        errors=errors,
        released_videos=released_videos,
        university=input_university,
    )
    if len(summary_df) > 0:
        summary_df.to_csv("/dev/stdout", index=False)
    else:
        print("No errors")

    print(f"Writing to directory: {output_dir}")
    pathmgr.mkdirs(output_dir)

    with pathmgr.open(os.path.join(output_dir, "errors.csv"), "w") as out_f:
        errors_df.to_csv(out_f, index=False)
    with pathmgr.open(os.path.join(output_dir, "summary.csv"), "w") as out_f:
        summary_df.to_csv(out_f, index=False)

    # NOTE: this is not the best solution, but we're likely going to be running
    # this once per python interpreter context
    stream_path_mgr = None



import math
import os
from collections import defaultdict
from typing import List, Tuple, Union

import h5py
import pandas as pd
import torch
from ego4d.research.clep.config import TrainConfig
from ego4d.research.dataset import LabelledFeatureDset
from torch import Tensor
from torch.utils.data import DataLoader, Dataset


def get_start_end_idx(t1, t2, feature_per_sec, nf):
    assert t2 >= 0
    x1 = min(
        max(0, math.floor(t1 * feature_per_sec)),
        nf - 1,
    )
    x2 = min(
        math.floor(t2 * feature_per_sec),
        nf - 1,
    )
    assert x2 >= x1
    return x1, x2 + 1


def _one_hot_encoding(n: int, clazzes: Union[int, List[int]]) -> torch.Tensor:
    result = torch.zeros(n)
    result[clazzes] = 1
    return result


def create_ego_charades_dset(
    config: TrainConfig, use_ego_sent: bool, ego_only: bool
) -> Tuple[
    Dataset,
    Tensor,
]:
    """
    Loads the ego charades dataset and returns a dataset and the associated
    sentence embeddings for each class (as a single tensor, where class_idx ==
    idx in the tensor).
    """
    val_df = pd.read_csv(config.pre_config.ego_charade.set_path)

    if ego_only is not None:
        if ego_only:
            val_df = val_df[val_df.egocentric == "Yes"]  # pyre-ignore
        else:
            val_df = val_df[val_df.egocentric == "No"]  # pyre-ignore

    val_df = val_df[~pd.isnull(val_df.actions)]

    sent = torch.load(
        os.path.join(
            config.pre_config.root_dir,
            config.pre_config.ego_charade.out_label_path,
        )
    )
    if use_ego_sent is not None:
        key = "sent_ego_fv" if use_ego_sent else "sent_non_ego_fv"
    else:
        key = "labels"

    sent_ordered = torch.stack([torch.tensor(fv) for fv in sent[key]])
    num_clazzes = len(sent_ordered)

    feature_hdf5_path = os.path.join(
        config.pre_config.root_dir,
        config.pre_config.ego_charade.out_path,
    )
    id_classes_pairs = []
    for row in val_df.itertuples():
        clazzes = []
        for x in row.actions.split(";"):
            clazzes.append(int(x.split(" ")[0].split("c")[1]))
        id_classes_pairs.append((row.id, _one_hot_encoding(num_clazzes, clazzes)))
    assert len(id_classes_pairs) == len(val_df)

    return LabelledFeatureDset(feature_hdf5_path, id_classes_pairs), sent_ordered


def create_kinetics_dset(
    config: TrainConfig,
) -> Tuple[
    Dataset,
    Tensor,
]:
    k400_config = config.pre_config.k400
    root = os.path.join(
        config.pre_config.root_dir,
        k400_config.root_dir,
    )

    sent_meta_path = os.path.join(root, k400_config.metadata_out_path)
    sent_features = torch.load(sent_meta_path)

    feature_hdf5_path = os.path.join(root, k400_config.viz_feature_path)
    label_name_to_idx = sent_features["label_name_to_idx"]
    sent_ordered = torch.stack([torch.tensor(fv) for fv in sent_features["label_fv"]])
    id_label_pairs = [
        (id, label_name_to_idx[label_name])
        for id, label_name in sent_features["labels"]
    ]
    return (
        LabelledFeatureDset(
            feature_hdf5_path, id_label_pairs, lambda x, _: x[0:].mean(0).squeeze()
        ),
        sent_ordered,
    )


class CCDset(Dataset):
    def __init__(
        self,
        config: TrainConfig,
    ):
        super().__init__()
        self.config = config
        self.viz_feature_path = os.path.join(
            config.pre_config.root_dir,
            config.pre_config.cc.hdf5_viz_path,
        )
        self.sent_feature_path = os.path.join(
            config.pre_config.root_dir,
            config.pre_config.cc.hdf5_sent_path,
        )
        self.viz_dset = h5py.File(self.viz_feature_path)
        self.sent_dset = h5py.File(self.sent_feature_path)

        self.keys = torch.load(config.pre_config.cc.meta_path)

    def __len__(self):
        return len(self.keys)

    def __getitem__(self, idx):
        k = self.keys[idx]
        v = self.viz_dset[k][0:]
        s = self.sent_dset[k][0:]
        return {
            "video": v,
            "text": s,
            "text_no_tag": s,
        }


class Ego4DCLEP(Dataset):
    def __init__(
        self,
        config: TrainConfig,
    ):
        super().__init__()

        self.narr_meta_path = os.path.join(
            config.pre_config.root_dir, config.pre_config.ego4d_narr.metadata_out_path
        )
        self.narr_meta = torch.load(self.narr_meta_path)
        self.config = config
        self.narr_feature_dir = os.path.join(
            config.pre_config.root_dir, config.pre_config.ego4d_narr.narration_out_dir
        )
        self.features = h5py.File(
            os.path.join(
                config.pre_config.root_dir,
                config.pre_config.ego4d_features.hdf5_path,
            )
        )
        uids = set(meta["uid"] for meta in self.narr_meta)
        assert len(uids - set(self.features.keys())) == 0, "not all features cached"
        self.narr_meta = [meta for meta in self.narr_meta if meta["uid"] in uids]

        t_by_uid = defaultdict(list)
        for x in self.narr_meta:
            t_by_uid[x["uid"]].append(x["ts"])

        self.betas = {
            uid: torch.mean(torch.tensor(v)[1:] - torch.tensor(v)[0:-1])
            for uid, v in t_by_uid.items()
        }
        self.alpha = torch.mean(torch.stack(list(self.betas.values())))

        old_len = len(self.narr_meta)
        self.narr_meta = [x for x in self.narr_meta if self.betas[x["uid"]] >= 1e-1]
        print(f"{old_len} -> {len(self.narr_meta)}")

    def __len__(self):
        return len(self.narr_meta)

    def __getitem__(self, idx):
        meta = self.narr_meta[idx]
        uid = meta["uid"]
        ts = meta["ts"]
        narr_idx = meta["idx"]

        # get txt feature
        txt_feature_path = os.path.join(self.narr_feature_dir, uid, f"{narr_idx}.pt")
        txt_feat = torch.load(txt_feature_path)

        offset = self.config.input_config.narration_width_sample_sec
        # offset = self.betas[uid] / (2*self.alpha)
        t1 = ts - offset
        t2 = ts + offset

        features = self.features[uid]
        start_idx, end_idx = get_start_end_idx(
            t1,
            t2,
            self.config.input_config.features_per_second,
            len(features),
        )
        features = features[start_idx:end_idx]
        v_feat = features.mean(0)  # aggregate

        return {
            "video": v_feat,
            "text": txt_feat["fv"],
            "text_no_tag": txt_feat["fv_no_tag"],
        }


def create_data_loader(dset, config: TrainConfig, shuffle=True):
    return DataLoader(
        dset,
        batch_size=config.batch_size,
        num_workers=config.num_workers,
        prefetch_factor=config.prefetch_factor,
        shuffle=shuffle,
    )



import functools
import os

from multiprocessing import Pool
from typing import List, Tuple

import h5py

import pandas as pd

import torch
import torchvision.transforms as T
from ego4d.features.config import FeatureExtractConfig, load_model
from ego4d.research.clep.config import CCPreprocessConfig, TrainConfig
from ego4d.research.clep.dataset import create_data_loader
from ego4d.research.clep.preprocess.common import get_language_model
from ego4d.research.common import batch_it, create_executor
from omegaconf import OmegaConf
from PIL import Image
from tqdm.auto import tqdm


class ImagePathDset(torch.utils.data.Dataset):
    def __init__(self, paths, transform):
        super().__init__()
        self.paths = paths
        self.transform = transform

    def __len__(self) -> int:
        return len(self.paths)

    def __getitem__(self, idx):
        path = self.paths[idx]
        img = Image.open(path).convert("RGB")
        img = self.transform(img)
        return {
            "img": img,
            "path": path,
        }


def preprocess_cc(config: TrainConfig, cc: CCPreprocessConfig):
    out_dir = config.pre_config.root_dir
    os.makedirs(out_dir, exist_ok=True)

    train_df = pd.read_csv(cc.in_path, sep="\t")

    examples = dict(zip(train_df.filepath, train_df.title))
    invalid_keys = []
    with Pool(cc.helper_workers) as pool:
        for path, x in tqdm(pool.imap(_get_fs, examples), total=len(examples)):
            if x == 0:
                invalid_keys.append(path)
    invalid_keys = set(invalid_keys)
    print(len(invalid_keys))
    examples = {k: v for k, v in examples.items() if k not in invalid_keys}

    feature_extract_config = OmegaConf.load(
        config.input_config.feature_extract_config_path
    )
    feature_extract_config.model_config.input_type = "image"  # not actually needed

    all_ex = list(examples.items())
    batches = batch_it(all_ex, cc.imgs_per_gpu)
    batches = [(idx, batch) for idx, batch in enumerate(batches)]

    # save off the keys to the dataset
    torch.save(
        [_get_key(p, idx) for idx, batch in batches for p, _ in batch], cc.meta_path
    )

    executor = create_executor(config.pre_config.slurm_config, len(batches))

    jobs = executor.map_array(
        functools.partial(
            _map_cc_batch,
            cc=cc,
            feature_extract_config=feature_extract_config,
        ),
        batches,
    )

    viz_path = os.path.join(out_dir, cc.hdf5_viz_path)
    with h5py.File(viz_path, "w") as out_f:
        for job in tqdm(jobs):
            res = job.result()
            for k, v in res.items():
                out_f.create_dataset(k, data=v)

    print("converting captions")
    jobs = executor.map_array(
        functools.partial(_map_cc_sent_batch, config=config, cc=cc),
        batches,
    )
    sent_path = os.path.join(out_dir, cc.hdf5_sent_path)
    with h5py.File(sent_path, "w") as out_f:
        for job in tqdm(jobs):
            res = job.result()
            for k, v in res.items():
                out_f.create_dataset(k, data=v)


def _get_fs(x):
    return x, os.path.getsize(x)


def _map_cc_batch(
    batch_idx, cc: CCPreprocessConfig, feature_extract_config: FeatureExtractConfig
):
    idx, batch = batch_idx
    paths = [x for x, _ in batch]
    viz_model = load_model(feature_extract_config, patch_final_layer=True)

    image_transform = T.Compose(
        [
            T.Resize(224),
            T.CenterCrop(224),
            T.ToTensor(),
            T.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
            T.Lambda(lambda x: x.unsqueeze(0).permute(1, 0, 2, 3)),
        ]
    )

    dset = ImagePathDset(paths, image_transform)
    dloader = create_data_loader(dset, cc, shuffle=False)  # pyre-ignore
    viz_model.eval()

    ret = {}
    with torch.no_grad():
        for xx in tqdm(dloader, total=len(dloader)):
            fv = viz_model({"video": xx["img"].cuda()})
            for p, f in zip(xx["path"], fv.cpu().numpy()):
                ret[_get_key(p, idx)] = f
    return ret


def _map_cc_sent_batch(
    batch_idx: Tuple[int, List[str]], config: TrainConfig, cc: CCPreprocessConfig
):
    idx, batch = batch_idx
    paths = [x for x, _ in batch]
    cap = [x for _, x in batch]

    sent_model = get_language_model(config)
    fvs = sent_model.encode(cap, device="cuda", show_progress_bar=True)
    ret = {}
    for p, f in zip(paths, fvs):
        ret[_get_key(p, idx)] = f
    return ret


def _get_key(p, idx):
    pn = p.replace("/", "_")
    return f"{idx}/{pn}"



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

from .s3path import bucket_and_key_from_path


def test_key_without_slashes():
    b, k = bucket_and_key_from_path("s3://bucket/key")
    assert b == "bucket"
    assert k == "key"


def test_key_with_slashes():
    b, k = bucket_and_key_from_path("s3://bucket/object/key")
    assert b == "bucket"
    assert k == "object/key"


def test_bucket_with_special_chars():
    b, k = bucket_and_key_from_path("s3://bucket-with.chars/object/key")
    assert b == "bucket-with.chars"
    assert k == "object/key"


def test_key_with_extension():
    b, k = bucket_and_key_from_path("s3://bucket-name/object/key.json")
    assert b == "bucket-name"
    assert k == "object/key.json"


def test_key_path_with_special_chars():
    b, k = bucket_and_key_from_path("s3://bucket-name/object$-/data/folder/key.json")
    assert b == "bucket-name"
    assert k == "object$-/data/folder/key.json"



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

"""
Verifies integrity of downloaded files
"""
from typing import Collection, List

from ego4d.cli.download import VideoOnDisk


def list_corrupt_files(downloads: Collection[VideoOnDisk]) -> List[VideoOnDisk]:
    """
    Returns a list of any downloaded files that appear corrupted.
    """
    return [d for d in downloads if _file_is_corrupt(d)]


def _file_is_corrupt(download: VideoOnDisk):
    if download.file_path.exists():
        return download.file_path.stat().st_size != download.s3_content_size_bytes

    return True



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

"""
A data object for storing user options, along with utilities for parsing input options
from command line flags and a configuration file.
"""
import argparse
import datetime
import json
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from ego4d.cli.universities import BUCKET_TO_UNIV, UNIV_TO_BUCKET


@dataclass
class ValidatedConfig:
    """
    Data object that stores validated user-supplied configuration options for a video
    download operation.
    """

    input_directory: str
    validate_all: bool
    metadata_folder: str
    input_university: str
    output_dir: str
    aws_profile_name: str
    num_workers: int
    expiry_time_sec: int
    version: str
    skip_mp4_check: bool
    released_video_path: Optional[str] = None


@dataclass
class Config:
    """
    Data object that stores the user-supplied configuration options for a video download
    operation.
    """

    input_directory: str
    validate_all: bool
    metadata_folder: str
    num_workers: int
    expiry_time_sec: int
    aws_profile_name: str
    version: str
    skip_mp4_check: bool
    released_video_path: Optional[str] = None
    input_university: Optional[str] = None
    output_dir: Optional[str] = None


def validate_config(cfg: Config) -> ValidatedConfig:
    """

    Args:
        cfg: A user-supplied configuration for the download operation
    Returns:
        A ValidatedConfig if all user-supplied options appear valid
    """

    def _maybe_fix_s3_folder(path: str) -> str:
        if path.startswith("s3") and not path.endswith("/"):
            return path + "/"
        return path

    if cfg.input_directory.startswith("s3://"):
        if cfg.input_university is not None:
            raise AssertionError("please do not provide a university name")

        bucket = cfg.input_directory.split("s3://")[1].split("/")[0]
        cfg.input_university = BUCKET_TO_UNIV[bucket]

    if cfg.input_university not in set(UNIV_TO_BUCKET.keys()):
        raise AssertionError(f"{cfg.input_university} is not a valid university")

    if cfg.aws_profile_name != "default":
        raise AssertionError(
            "aws profile!=default: other profiles not currently supported"
        )

    if cfg.output_dir is None:
        time_now = f"{datetime.datetime.now().strftime('%d%m%y_%H%M%S')}"
        cfg.output_dir = f"s3://ego4d-consortium-sharing/internal/validation/{cfg.input_university}/{time_now}/"  # noqa
        print(f"Using output directory: {cfg.output_dir}")

    if cfg.metadata_folder is None:
        cfg.metadata_folder = (
            "ego4d/internal/validation/standard_metadata/egoexo"
            if cfg.version == "egoexo"
            else "ego4d/internal/validation/standard_metadata/ego4d/"
        )

    return ValidatedConfig(
        input_directory=_maybe_fix_s3_folder(cfg.input_directory),
        validate_all=bool(cfg.validate_all),
        metadata_folder=_maybe_fix_s3_folder(cfg.metadata_folder),
        released_video_path=cfg.released_video_path,
        input_university=cfg.input_university,
        output_dir=_maybe_fix_s3_folder(cfg.output_dir),
        aws_profile_name=cfg.aws_profile_name,
        num_workers=cfg.num_workers,
        expiry_time_sec=cfg.expiry_time_sec,
        version=cfg.version,
        skip_mp4_check=cfg.skip_mp4_check,
    )


def config_from_args(args=None) -> Config:
    """
    Parses command line flags and returns a Config object with corresponding values from
    the flags.
    """
    # Parser for a configuration file
    json_parser = argparse.ArgumentParser(
        description="Command line tool to download Ego4D datasets from Amazon S3"
    )

    json_parser.add_argument(
        "--config_path",
        type=Path,
        help="Local path to a config JSON file. If specified, the flags will be read "
        "from this file instead of the command line.",
    )

    args, remaining = json_parser.parse_known_args(args=args)

    # Parser for command line flags other than the configuration file
    flag_parser = argparse.ArgumentParser(add_help=False)

    # required_flags = {"output_directory"}
    flag_parser.add_argument(
        "-i",
        "--input_directory",
        help="The S3 path where the university's video metadata is stored",
    )

    flag_parser.add_argument(
        "-a",
        "--all",
        default=False,
        help="validate all files in S3",
        dest="validate_all",
    )
    flag_parser.add_argument(
        "-mf",
        "--metadata_folder",
        help="The S3 path where the device/component_type/scenario metadata is stored",
        default=None,
    )
    flag_parser.add_argument(
        "-rp",
        "--released_video_path",
        help="The path where released_videos file is stored",
        default=None,
    )
    flag_parser.add_argument(
        "-u",
        "--input_university",
        help="The university name we're checking data on",
        default=None,
    )
    flag_parser.add_argument(
        "-o",
        "--output_dir",
        help="output directory",
    )
    flag_parser.add_argument(
        "-nw",
        "--num_workers",
        type=int,
        help="number of workers",
        default=10,
    )
    flag_parser.add_argument(
        "-expiry",
        "--expiry_time_sec",
        type=int,
        help="default expiry for presigned URLs (in seconds)",
        default=3600,
    )
    flag_parser.add_argument(
        "--aws_profile_name",
        help="Defaults to 'default'. Specifies the AWS profile name from "
        "~/.aws/credentials to use for the download",
        default="default",
    )
    flag_parser.add_argument(
        "--version",
        help="EgoExo validation or Ego4D validation?",
        default="egoexo",
    )
    flag_parser.add_argument(
        "--skip_mp4_check",
        default=False,
        help="skip checking mp4 files with ffprobe",
        action="store_true",
    )

    # Use the values in the config file, but set them as defaults to flag_parser so they
    # can be overridden by command line flags
    if args.config_path:
        with open(args.config_path.expanduser()) as f:
            config_contents = json.load(f)
            flag_parser.set_defaults(**config_contents)

    parsed_args = flag_parser.parse_args(remaining)

    flags = {k: v for k, v in vars(parsed_args).items()}

    # Note: Since the flags from the config file are being used as default argparse
    # values, we can't set required=True. Doing so would mean that the user couldn't
    # leave them unspecified when invoking the CLI.
    config = Config(**flags)

    return config



import os
import random
import time
import traceback
from dataclasses import dataclass
from typing import Any, List, Optional

import boto3
import boto3.session
import botocore.client as bclient
import botocore.exceptions
from iopath import PathManager

S3Client = bclient.BaseClient


@dataclass
class S3FileDesc:
    basename: str
    path: str
    size: int


def _get_location(bucket_name: str) -> str:
    client = boto3.client("s3")
    response = client.get_bucket_location(Bucket=bucket_name)
    return response["LocationConstraint"]


def get_config(
    num_workers: int = 10,
    connect_timeout: int = 180,
    max_attempts: int = 3,
    region_name: Optional[str] = None,
):
    config = bclient.Config(
        region_name=region_name,
        connect_timeout=connect_timeout,
        max_pool_connections=num_workers,
        retries={"mode": "standard", "max_attempts": max_attempts},
    )
    return config


def get_session(profile: Optional[str]):
    return boto3.session.Session(profile_name=profile)


def get_client(
    bucket_name: str,
    num_workers: int = 10,
    connect_timeout: int = 180,
    max_attempts: int = 3,
    profile: Optional[str] = None,
) -> S3Client:
    session = get_session(
        profile=profile,
    )
    return session.client(
        "s3",
        config=get_config(
            region_name=_get_location(bucket_name),
            num_workers=num_workers,
            connect_timeout=connect_timeout,
            max_attempts=max_attempts,
        ),
    )


def get_resource(
    profile: str,
    num_workers: int = 10,
    connect_timeout: int = 180,
    max_attempts: int = 3,
) -> Any:
    session = get_session(profile=profile)
    return session.resource(
        "s3",
        config=get_config(
            region_name=None,
            num_workers=num_workers,
            connect_timeout=connect_timeout,
            max_attempts=max_attempts,
        ),
    )


class StreamPathMgr:
    def __init__(self, expiration_sec: int = 7200):
        self.clients = {}
        self.expr_sec = expiration_sec
        self.cached_paths = {}

    def open(self, path: str) -> str:
        if path in self.cached_paths:
            return self.cached_paths[path]

        if path.startswith("s3"):
            temp = path.split("s3://")[1].split("/")
            bucket_name = temp[0]
            object_name = "/".join(temp[1:])

            client = None
            if bucket_name not in self.clients:
                self.clients[bucket_name] = get_client(bucket_name)

            client = self.clients[bucket_name]
            ret = client.generate_presigned_url(
                "get_object",
                Params={"Bucket": bucket_name, "Key": object_name},
                ExpiresIn=self.expr_sec,
            )
            self.cached_paths[path] = ret
            return ret
        else:
            return path


def ls_relative(path: str, pathmgr: PathManager) -> List[str]:
    files = pathmgr.ls(path)
    return [f.split("/")[-1] for f in files]


# TOOD(miguelmartin): removeme, duplicate function
def exp_backoff(max_sleep_time_sec=1200, base=2):  # pyre-ignore
    def decorator(func):  # pyre-ignore
        assert base > 1

        def wrapper(*args, **kwargs):  # pyre-ignore
            sleep_t = base
            while True:
                try:
                    ret = func(*args, **kwargs)
                    return ret
                except Exception as e:
                    print(
                        f"WARN: exp_backoff retrying...: {func.__name__}, {traceback.format_exc()}"
                    )
                    time.sleep(sleep_t / base + sleep_t * random.random())
                    sleep_t *= base
                    if sleep_t > max_sleep_time_sec:
                        print(
                            f"ERROR: exp_backoff FAILED: {func.__name__}, {traceback.format_exc()}"
                        )
                        raise e
                    continue

        return wrapper

    return decorator


# NOTE:
# iopath / pathmgr not used as we cannot get an S3 Object for sizing with it
class S3Downloader:
    def __init__(self, profile, num_workers=10, callback=None):
        self.clients = {}
        self.resources = get_resource(profile=profile, num_workers=num_workers)
        self.profile = profile
        self.callback = callback

    def ls(self, path, recursive=False, max_keys=-1) -> List[S3FileDesc]:
        assert path.startswith("s3://")
        temp = path.split("s3://")[1].split("/")
        bucket_name = temp[0]
        object_name = "/".join(temp[1:])
        if bucket_name not in self.clients:
            self.clients[bucket_name] = get_client(
                bucket_name,
                profile=self.profile,
            )  # pyre-ignore

        delim = "/" if not recursive else ""
        client = self.clients[bucket_name]
        if max_keys < 0:
            paginator = client.get_paginator("list_objects_v2")
            pages = paginator.paginate(
                Bucket=bucket_name, Prefix=object_name, Delimiter=delim
            )
            return [
                S3FileDesc(
                    basename=os.path.basename(f["Key"]),
                    path=f"s3://{bucket_name}/{f['Key']}",
                    size=f["Size"],
                )
                for page in pages
                for f in page.get("Contents", [])
            ]
        else:
            assert max_keys <= 1000
            ls_result = client.list_objects_v2(
                Bucket=bucket_name,
                Prefix=object_name,
                MaxKeys=max_keys,
                Delimiter="/",
            )
            if ls_result["KeyCount"] == 0:
                return []
            return [
                S3FileDesc(
                    basename=os.path.basename(f["Key"]),
                    path=f"s3://{bucket_name}/{f['Key']}",
                    size=f["Size"],
                )
                for f in ls_result["Contents"]
            ]

    def copy(self, path: str, out_path: str):
        self.obj(path).download_file(out_path, Callback=self.callback)

    def file_desc(self, path: str) -> Optional[S3FileDesc]:
        o = self.obj(path)
        if o is None:
            return None
        try:
            file_size = o.content_length
            ret = S3FileDesc(
                basename=os.path.basename(o.key),
                path=path,
                size=file_size,
            )
        except botocore.exceptions.ClientError as e:
            # If a client error is thrown, then check that it was a 404 error.
            # If it was a 404 error, then the file does not exist.
            error_code = int(e.response["Error"]["Code"])
            if error_code == 404:
                return None
            else:
                raise e
        return ret

    def obj(self, path):
        assert path.startswith("s3://")
        temp = path.split("s3://")[1].split("/")
        bucket_name = temp[0]
        object_name = "/".join(temp[1:])

        try:
            return self.resources.Object(bucket_name, object_name)
        except botocore.exceptions.ClientError as e:
            # If a client error is thrown, then check that it was a 404 error.
            # If it was a 404 error, then the file does not exist.
            error_code = int(e.response["Error"]["Code"])
            if error_code == 404:
                return None
            else:
                raise e



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved.

"""
Command line tool to download Ego4D datasets.

Examples:
      python -m ego4d.cli.cli \
        --version="v2_1" \
        --datasets full_scale annotations \
        --output_directory="~/ego4d_data"
"""
import logging
from pathlib import Path
from typing import List

import boto3
from ego4d.cli.config import (
    Config,
    config_from_args,
    DATASET_PRIMARY,
    DATASETS_VIDEO,
    validate_config,
)
from ego4d.cli.download import (
    create_download_directory,
    create_output_directory,
    download_all,
    FileToDownload,
    filter_already_downloaded,
    list_corrupt_files,
    list_videos_for_download,
    load_version_file,
    save_version_file,
    upsert_version,
)
from ego4d.cli.manifest import (
    download_manifest_for_version,
    download_metadata,
    print_datasets,
)
from ego4d.cli.progressbar import DownloadProgressBar
from tqdm import tqdm


def main_cfg(cfg: Config) -> None:

    validated_cfg = validate_config(cfg)

    # This service resource in the default session will be used for general light-weight
    # requests on the main thread, such as downloading the video manifests and getting
    # S3 object metadata
    s3 = boto3.session.Session(profile_name=validated_cfg.aws_profile_name).resource(
        "s3"
    )

    if cfg.video_uids and all(x not in DATASETS_VIDEO for x in validated_cfg.datasets):
        logging.error(
            "ERROR: video_uids specified for non-video datasets (and will be ignored)"
        )

    if validated_cfg.list_datasets:
        print_datasets(validated_cfg.version, s3)
        return

    print(f"Datasets to download: {validated_cfg.datasets}")

    # TODO: Handle not output dir
    output_path = create_output_directory(validated_cfg)
    print(f"Download Path: {output_path}")

    # Download the primary metadata to the root directory
    if cfg.metadata:
        metadata_path = download_metadata(
            validated_cfg.version,
            validated_cfg.output_directory,
            s3,
        )
        if not metadata_path:
            logging.error("ERROR: Primary Metadata Download Failed")
        else:
            print(f"Ego4D Metadata: {metadata_path}")

    # Download the manifest to the root directory
    if cfg.manifest:
        toplevel_manifest_path = download_manifest_for_version(
            validated_cfg.version,
            DATASET_PRIMARY,
            validated_cfg.output_directory,
            s3,
        )
        if not toplevel_manifest_path:
            logging.error("ERROR: Primary Manifest Download Failed")
            print("ABORT: Primary Manifest Download Failed")
            return

        print(f"Manifest downloaded: {toplevel_manifest_path}")

    downloads: List[FileToDownload] = []
    print("Checking requested datasets and versions...")
    for dataset in validated_cfg.datasets:
        if dataset != dataset.lower():
            dataset = dataset.lower()
            print(f"ERROR: Please specify datasets as lower case.  Assuming: {dataset}")

        download_path = create_download_directory(validated_cfg, dataset)
        print(
            f"Created download directory for version '{validated_cfg.version}' of "
            f"dataset: '{dataset}' at: {download_path}"
        )

        if cfg.manifest_override_path:
            p = Path(cfg.manifest_override_path).expanduser()
            print(f"Overriding {dataset} manifest path: {p}")
            if not p.exists():
                logging.error(f"Manifest override doesn't exists: {p}")
            manifest_path = str(p)
        else:
            manifest_path = download_manifest_for_version(
                validated_cfg.version, dataset, download_path, s3
            )

        version_entries = load_version_file(download_path)

        if validated_cfg.video_uids:
            print(
                "Only downloading a subset of the video files because the "
                "'video_uids' flag has been set on the command line or in the config "
                f"file. A total of {len(validated_cfg.video_uids)} video files will "
                f"be downloaded.\n"
            )
        to_download = list_videos_for_download(validated_cfg, dataset, manifest_path)
        downloads.extend(
            [FileToDownload.create(video, download_path) for video in to_download]
        )

    print("Retrieving object metadata from S3...")
    cnt_invalid = 0
    for x in tqdm(downloads, unit="object"):
        if not x.s3_object_key:
            cnt_invalid += 1
            continue
        x.s3_object = s3.Object(x.s3_bucket, x.s3_object_key)

    if cnt_invalid == len(downloads):
        print("ABORT: All S3 Objects Invalid")
        return
    elif cnt_invalid > 0:
        logging.error(
            f"{cnt_invalid}/{len(downloads)} invalid S3 downloads will be ignored"
        )

    print("Checking if latest file versions are already downloaded...")
    active_downloads = filter_already_downloaded(
        downloads,
        version_entries,
        bypass_version_check=validated_cfg.bypass_version_check,
        skip_s3_checks=validated_cfg.skip_s3_checks,
    )

    missing = [x for x in downloads if not x.s3_exists]
    if len(missing) > 0:
        logging.error(
            f"{len(missing)}/{len(downloads)} missing S3 downloads will be ignored"
        )

    if len(active_downloads) == 0:
        print(
            "The latest versions of all requested videos already exist in the output "
            "directories under:\n"
            f"{validated_cfg.output_directory}"
        )
        exit(0)

    assert all(x.s3_object for x in active_downloads)

    total_size_bytes = None
    if not validated_cfg.skip_s3_checks:
        total_size_bytes = sum(
            x.s3_object.content_length for x in active_downloads if x.s3_object
        )
        if total_size_bytes == 0:
            print(
                "The latest versions of all requested videos already exist in the output "
                "directories under:\n"
                f"{validated_cfg.output_directory}"
            )
            exit(0)

        expected_gb = total_size_bytes / 1024 / 1024 / 1024

    if validated_cfg.assume_yes:
        if expected_gb:
            print(f"Downloading {len(active_downloads)} files..")
        else:
            print(f"Downloading {expected_gb:.4f} GB..")
    else:
        confirm = None
        while confirm is None:
            if validated_cfg.skip_s3_checks:
                response = input(
                    f"Number of files to download is: {len(active_downloads)}\n"
                    f"Run without --skip-s3-checks for a size estimate.\n"
                    f"Do you want to start the download? ([y]/n) "
                )
            else:
                response = input(
                    f"Expected size of downloaded files is "
                    f"{expected_gb:.4f} GB. "
                    f"Do you want to start the download? ([y]/n) "
                )
            if response.lower() in ["yes", "y", ""]:
                confirm = True
            elif response.lower() in ["no", "n"]:
                print("Aborting the download operation.")
                exit(0)
            else:
                continue

    progress = DownloadProgressBar(total_size_bytes=total_size_bytes)

    files = download_all(
        active_downloads,
        version_entries,
        aws_profile_name=validated_cfg.aws_profile_name,
        callback=progress.update,
        save_callback=lambda: save_version_file(version_entries, download_path),
    )

    print("Checking file integrity...")
    corrupted = list_corrupt_files(files)

    if corrupted:
        msg = f"ERROR: {len(corrupted)} files failed download: "
        for x in corrupted:
            msg += f"\t{x.uid}\n"
        logging.error(msg)
        # TODO: retry these downloads?

    # for x in active_downloads:
    #     upsert_version(x, version_entries)

    # One additional save to confirm
    save_version_file(version_entries, download_path)


def main() -> None:
    config = config_from_args()
    if not config:
        return
    main_cfg(config)


if __name__ == "__main__":
    main()



import copy

import hydra
import torch
import torch.nn.functional as F

from ego4d.features.config import FeatureExtractConfig, load_model
from ego4d.research.clep.config import TrainConfig
from ego4d.research.clep.dataset import create_data_loader, create_kinetics_dset
from ego4d.research.clep.utils import charades_map
from omegaconf import OmegaConf
from torch.nn import Identity

from tqdm.auto import tqdm


# taken from: https://github.com/mlfoundations/open_clip/blob/main/src/training/zero_shot.py#L29
def accuracy(output, target, topk=(1,)):
    pred = output.topk(max(topk), 1, True, True)[1].t()
    correct = pred.eq(target.view(1, -1).expand_as(pred))
    return [
        float(correct[:k].reshape(-1).float().sum(0, keepdim=True).cpu().numpy())
        for k in topk
    ]


def eval_multi_class_classification(model, classifier, loader, device=None):
    pred_arr = []
    gt_arr = []
    with torch.no_grad():
        for x, target in tqdm(loader):
            assert x.shape[0] == 1

            if device is not None:
                x = x.to(device)
                target = target.to(device)

            v = model(x)

            # https://github.com/mlfoundations/open_clip/blob/main/src/training/zero_shot.py#L49
            if isinstance(classifier, torch.Tensor):
                v = F.normalize(v, dim=-1)
                logits = v @ classifier
            else:
                logits = classifier(v)

            pred = torch.nn.Softmax(dim=1)(logits.mean(1))
            pred_arr.append(pred.squeeze().cpu())
            gt_arr.append(target.squeeze().cpu())

    # res = mAP(torch.stack(pred_arr).numpy(), torch.stack(gt_arr).numpy())
    res = charades_map(torch.stack(pred_arr).numpy(), torch.stack(gt_arr).numpy())
    return {
        "mAP": res[0] * 100,
    }


def eval_classification(model, classifier, loader, device=None, avg_logits=False):
    with torch.no_grad():
        acc1, acc5, n = 0.0, 0.0, 0
        for x, target in tqdm(loader):
            if avg_logits:
                assert x.shape[0] == 1

            if device is not None:
                x = x.to(device)
                target = target.to(device)

            v = model(x)

            # https://github.com/mlfoundations/open_clip/blob/main/src/training/zero_shot.py#L49
            if isinstance(classifier, torch.Tensor):
                v = F.normalize(v, dim=-1)
                logits = v @ classifier
            else:
                logits = classifier(v)

            if avg_logits:
                a1, a5 = accuracy(logits.mean(1), target, topk=(1, 5))
            else:
                a1, a5 = accuracy(logits, target, topk=(1, 5))

            # This is to test correctness
            # hot1 = torch.zeros(x.shape[0], 400)
            # for i in range(x.shape[0]):
            #     hot1[i, target[i]] = 1.0
            # a1, a5 = accuracy(hot1.to(self.device), target, topk=(1, 5))

            acc1 += a1
            acc5 += a5
            n += x.shape[0]

    acc1 /= n
    acc5 /= n
    return {
        "acc1": 100 * acc1,
        "acc5": 100 * acc5,
    }


def eval_k400_on_features(
    config: TrainConfig, feature_extract_config: FeatureExtractConfig
):
    # NOTE: only works for omnivore right now
    dset, _ = create_kinetics_dset(config)
    config = copy.deepcopy(config)
    config.batch_size = 1
    val_loader = create_data_loader(dset, config)

    omni_model = load_model(feature_extract_config, patch_final_layer=False)
    omni_model = omni_model.cuda()
    classifier = omni_model.model.heads.video[1]  # linear layer
    assert isinstance(classifier, torch.nn.Linear)

    model = Identity()
    classifier.eval()

    return eval_classification(model, classifier, val_loader, "cuda")


@hydra.main(config_path="configs", config_name=None)
def main(config: TrainConfig):
    feature_extract_config = OmegaConf.load(
        config.input_config.feature_extract_config_path
    )
    res = eval_k400_on_features(config, feature_extract_config)
    print(res)


if __name__ == "__main__":
    main()  # pyre-ignore
